"""
Utility functions for doing low-level, general-purpose EBML reading and writing.

Note: Most of the functionality that was here is now part of ebmlite. This
    module now contains little besides the 'Attribute' handling functions.

Created on Dec 10, 2013

:author: dstokes

"""

from collections import OrderedDict
import datetime
import time

from ebmlite import loadSchema


# ==============================================================================
# 
# ==============================================================================

def verify(data, schema=None):
    """ Basic sanity-check of data validity. If the data is bad an exception
        will be raised. The specific exception varies depending on the problem
        in the data.
        
        :keyword schema: The full module name of the EBML schema.
        :return: `True`. Any problems will raise exceptions.
    """
    if schema is None:
        schema = loadSchema('mide_ide.xml')
    return schema.verify(data)


# ==============================================================================
# 
# ==============================================================================

def decode_attributes(data, withTypes=False):
    """ Convert a set of Attributes (as a list of dictionaries containing an
        `AttributeName` and one of the Attribute value elements (`IntAttribute`,
        `FloatAttribute`, etc.) to a proper dictionary. Attributes are tagged
        as 'multiple,' so they become lists when the EBML is parsed.
    """
    result = OrderedDict()
    for atts in data:
        k = atts.pop('AttributeName', None)
        if k is None:
            continue
        t, v = list(atts.items())[0]
        if withTypes:
            att = (v, t)
        else:
            att = v
        result.setdefault(k, []).append(att)
    return result

    
def encode_attributes(data):
    """ Construct a set of `Attribute` dicts from a dictionary or list of
        tuples/lists. Each value should be either a simple data type or a
        tuple containing the value and the specific value element name
        (IntAttribute, FloatAttribute, etc.). The value element type will 
        otherwise be inferred.
    """
    datetimeTypes = (datetime.datetime, datetime.date)
    attTypes = ((str, 'UnicodeAttribute'),
                (bytes, 'StringAttribute'),
                (float, 'FloatAttribute'),
                (int, 'IntAttribute'),
                (time.struct_time, 'DateAttribute'),
                (datetimeTypes, 'DateAttribute'))
    
    if isinstance(data, dict):
        data = list(data.items())
    
    result = []
    elementType = None

    for d in data:
        if isinstance(d[1], (tuple, list)) and not isinstance(d[1], time.struct_time):
            k = d[0]
            v, elementType = d[1]
        else:
            k, v = d
            for t, elementType in attTypes:
                if isinstance(v, t):
                    break
                elementType = 'BinaryAttribute'
        
        if elementType == 'DateAttribute':
            if isinstance(v, (int, float)):
                v = int(v)
            elif isinstance(data, datetimeTypes):
                v = int(time.mktime(data.timetuple()))
            elif isinstance(data, time.struct_time):
                v = int(time.mktime(v))
                
        result.append({'AttributeName': k, elementType: v})

    return result
    

def build_attributes(data):
    """ Construct `Attribute` EBML from dictionary or list of key/value pairs. 
        Each value should be either a simple data type or a tuple containing 
        the value and the specific value element name (`IntAttribute`, 
        `FloatAttribute`, etc.). The value element type will otherwise be 
        inferred.
    """
    mideSchema = loadSchema('mide_ide.xml')
    return mideSchema['Attribute'].encode(encode_attributes(data))



'''
MATLAB .MAT file exporting.
'''

from datetime import datetime
from glob import glob
import os.path
import string
import struct

import logging
logger = logging.getLogger('idelib-archive')
logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s")

# NOTE: 64 bit Scipy is unstable; avoid using it for now (v0.13.2, 12/2014).
# from scipy.io.matlab import mio5_params as MP

class MP(object):
    """ MAT data and matrix types. """
    miINT8 =            0x01
    miUINT8 =           0x02
    miINT16 =           0x03
    miUINT16 =          0x04
    miINT32 =           0x05
    miUINT32 =          0x06
    miSINGLE =          0x07
    miDOUBLE =          0x09
    miINT64 =           0x0c
    miUINT64 =          0x0d
    miMATRIX =          0x0e
    miCOMPRESSED =      0x0f
    miUTF8 =            0x10
    miUTF16 =           0x11
    miUTF32 =           0x12
    
    mxCELL_CLASS =      0x01
    mxSTRUCT_CLASS =    0x02
    mxOBJECT_CLASS =    0x03
    mxCHAR_CLASS =      0x04
    mxSPARSE_CLASS =    0x05
    mxDOUBLE_CLASS =    0x06
    mxSINGLE_CLASS =    0x07
    mxINT8_CLASS =      0x08
    mxUINT8_CLASS =     0x09
    mxFUNCTION_CLASS =  0x10
    mxINT16_CLASS =     0x0a
    mxUINT16_CLASS =    0x0b
    mxINT32_CLASS =     0x0c
    mxUINT32_CLASS =    0x0d
    mxINT64_CLASS =     0x0e
    mxUINT64_CLASS =    0x0f
    mxOPAQUE_CLASS =    0x11
    mxOBJECT_CLASS_FROM_MATRIX_H = 0x12


#===============================================================================
# 
#===============================================================================

def splitNum(name, digits=string.digits):
    """ Split a string that ends with a number between the 'body' of the
        string and its numeric suffix. Both parts are returned as strings.
        
        :param name: The string to split.
        :keyword digits: The set of numeric characters. Defaults to 0-9.
        :return: A tuple containing the base name and the numeric suffix.
    """
    base = name.rstrip(digits)
    return base, name[len(base):]
   

def serialFilename(basename, numDigits=2, minNumber=1, inc=1, pad='_'):
    """ Generate a new filename with an incremented numeric suffix.

        :param basename: The name to make unique.
        :keyword numDigits: The minimum number of digits to use when adding a
            new number to a name. Used if there are no existing files.
        :keyword numNumber: The minimum serial number, if no file currently
            exists.
        :keyword inc: The serial number increment.
        :keyword pad: A string to appear between the base name and the number.
        :return: A unique filename.
    """
    base, ext = os.path.splitext(basename)
    existing = glob(base.rstrip(string.digits)+'*'+ext)
    existing = [x for x in existing if os.path.splitext(x)[0][-1].isdigit()]
    if len(existing) > 0:
        existing.sort()
        lastName, lastNum = splitNum(os.path.splitext(existing[-1])[0])
        numDigits = len(existing[-1]) - len(lastName) - len(ext)
        lastNum = int(lastNum)+inc
    else:
        lastNum = minNumber
    return '%s%s%s%s' % (base, pad, str(lastNum).rjust(numDigits, '0'), ext)
    

# def dump8(data, start=0, end=None):
#     """ Debugging tool. Prints data as 8 columns of hex. """
#     end = len(data) if end is None else end
#     for i in range(start,end,8):
#         print '%s:' % (str(i).rjust(4)), 
#         try:
#             for j in range(8):
#                 print "%02x" % (ord(data[i+j])),
#         except IndexError:
#             print "(end)",
#             break
#         print
# 
# def hexdump(data):
#     """ Debugging tool. Dumps a string or a collection of numbers as hex. """
#     if isinstance(data, (str, bytes, bytearray)):
#         a = ["%02x" % ord(c) for c in data]
#     else:
#         a = ["%02x" % c for c in data]
#     return ' '.join(a)

# MATLAB reserved words. 
# TODO: Add standard functions as well
RESERVED_WORDS = ('break', 'case', 'catch','continue', 'else', 'elseif', 'end',
                  'for', 'function', 'global', 'if', 'otherwise', 'persistent',
                  'return', 'switch', 'try', 'while')

def sanitizeName(s, validChars=string.ascii_letters+string.digits+'_',
                 prefix="v", reservedWords=RESERVED_WORDS):
    """ Convert an arbitrary string into a valid MATLAB variable name.
    
        :keyword validChars: A string of all valid characters
    """
    s = s.strip().encode('ascii', 'replace')
    result = [c if c in validChars else '_' for c in s.strip()]
    result = ''.join(result).strip('_ ')
    if result[0].isdigit() or result in reservedWords:
        result = prefix + result.title()
    return result.replace('__','_').replace('__','_').encode('utf8')


#===============================================================================
# 
#===============================================================================

class MatStream(object):
    """
    """
    MAX_LENGTH = (2**31)-9 # Accounts for data being rounded to the next x8.
    MAX_SIZE = int(MAX_LENGTH * .95) # scale back by 5%, just to be certain
    DEFAULT_HEADER = "MATLAB 5.0 MAT-file MIDE IDE to MAT"
    
    # Map MATLAB types to the struct formatting character
    typeFormatChars = {
        MP.miINT8:   'b',
        MP.miUINT8:  'B',
        MP.miINT16:  'h',
        MP.miUINT16: 'H',
        MP.miINT32:  'i',
        MP.miUINT32: 'I',
        MP.miINT64:  'q',
        MP.miUINT64: 'Q',
        MP.miUTF8:   'c',
        MP.miSINGLE: 'f',
        MP.miDOUBLE: 'd',
    }
    
    # Map data types to matrix types
    classTypes = {
        MP.miINT8:   MP.mxINT8_CLASS,
        MP.miUINT8:  MP.mxUINT8_CLASS,
        MP.miINT16:  MP.mxINT16_CLASS,
        MP.miUINT16: MP.mxUINT16_CLASS,
        MP.miINT32:  MP.mxINT32_CLASS,
        MP.miUINT32: MP.mxUINT32_CLASS,
        MP.miINT64:  MP.mxINT64_CLASS,
        MP.miUINT64: MP.mxUINT64_CLASS,
        MP.miUTF8:   MP.mxCHAR_CLASS,
        MP.miSINGLE: MP.mxSINGLE_CLASS,
        MP.miDOUBLE: MP.mxDOUBLE_CLASS,
    }
    
    intPack = struct.Struct('I')


    def __init__(self, filename, doc=None, msg=None, serialize=True, 
                 maxFileSize=MAX_SIZE, timeScalar=1, writeCal=False,
                 calChannels=None, writeStart=False, writeInfo=False):
        """ Constructor. Create a new .MAT file.
        
            :param filename: The name of the new file, or `None`.
            :keyword doc: The `idelib-archive.dataset.Dataset` from which to export.
            :keyword msg: The message string to appear at the start of the
                MAT file. The first 4 bytes must be non-zero.
            :keyword serialize: If `True`, start the first file with a 
                number. Successive files are always numbered.
            :keyword maxFileSize: The maximum size of each exported file.
                Must not exceed the maximum size allowed for MATs.
            :keyword timeScalar: Timestamp scaling factor.
            :keyword writeCal: If `True`, write calibration data to each
                exported file. If `"channel"`, write calibration by channel
                name, ignoring any transforms not being used.
            :keyword writeStart: If `True`, write the doc's start time.
            :keyword writeInfo: If `True`, write doc/recorder info.
        """
        if msg is None:
            msg = self.DEFAULT_HEADER if doc is None else self.makeHeader(doc)
        self.basename = filename
        self.filename = filename
        self.msg = msg.encode('utf8') or self.DEFAULT_HEADER
        self.stream = None
        self.timeScalar = timeScalar
        self.doc = doc
        if doc is not None:
            self._writeCal = writeCal and len(doc.transforms) > 0
            self._calPerChannel = "channel" in str(writeCal).lower()
            self._writeInfo = writeInfo
            self.startTime = writeStart and doc.sessions[0].utcStartTime
        else:
            self._writeCal = self._writeInfo = self._writeStart = False
        self.calChannels = calChannels

        # MATLAB identifies the file as level 4 if the first byte is 0.
        if '\x00' in self.msg[:4]:
            self.msg = self.DEFAULT_HEADER
            
        self.maxFileSize = min(self.MAX_SIZE, self.next8(maxFileSize))
        
        # Default array parameters. Typically set by startArray().
        self._inArray = False
        self.arrayName = None
        self.arrayMType = None
        self.arrayDType = MP.miDOUBLE
        self.arrayFlags = 0
        self.arrayNoTimes = False
        self.arrayHasTimes = False
        self.arrayColNames = None
        
        self.exportedFiles = []
        
        if filename is not None:
            self.newFile(serialize)
        else:
            self.stream = None
        

    def newFile(self, serialize=True, offset=None):
        """ Used internally. Wraps file creation.
        """
        if self.stream is not None:
            self.close()
            
        if serialize:
            self.filename = serialFilename(self.basename)
        else:
            self.filename = self.basename
        self.stream = open(self.filename, 'wb')
        self._write = self.stream.write
        self._seek = self.stream.seek

        self._write(struct.pack('116s II H 2s', self.msg, 0, 0, 0x0100, 'IM'))
        
        if self._writeInfo:
            self.writeRecorderInfo(self.doc.recorderInfo)
        if self._writeCal:
            if self._calPerChannel:
                self.writeCalPerChannel(self.doc)
            else:
                self.writeCalibration(self.doc.transforms)
        if self.startTime:
            self.writeValue('start_time_utc', self.startTime, MP.miINT64)
        
        self._inArray = False
        self.arrayName = None
        self.arrayColNames = None
        self.arrayStartTime = None
        
        self.exportedFiles.append(self.filename)


    def checkFileSize(self, size, pad=0):
        """ Check if adding data of the given size will exceed the maximum    
            file size; start a new file if it will. Also creates a variable
            for the current array's start time (if provided).
        """
        if self.next8(self.stream.tell() + size + pad) >= self.maxFileSize:
#             print "checkFileSize"
            inArray = self._inArray
            if inArray:
                colNames = self.arrayColNames
                baseName = self.arrayBaseName
                dtype = self.arrayDType
                mtype = self.arrayMType
                noTimes = self.arrayNoTimes
                hasTimes = self.arrayHasTimes
                num = self.arrayNumber + 1
                cols = self.numCols if self.arrayNoTimes else self.numCols - 1
            self.newFile()
            if inArray:
                self.startArray(baseName, cols, arrayNumber=num, 
                                colNames=colNames, mtype=mtype, dtype=dtype, 
                                noTimes=noTimes, hasTimes=hasTimes)


    @property
    def closed(self):
        return True if self.stream is None else self.stream.closed


    @classmethod
    def next8(cls, num):
        """ Return the next multiple of 8; elements within a MAT file must be
            aligned to 64 bits.
        """
        if num % 8 == 0:
            return num
        return 8*(int(num/8)+1)


    def pack(self, fmt, args, dtype=MP.miUINT32):
        """ Write data to the file, proceeded by the type and size (aligned to
            64 bits). Used internally.
            
            :param fmt: A struct formatting string for the data.
            :param args: Arguments for the struct formatting string (i.e. data).
            :keyword dtype: The type for the data size element.
        """
        dataSize = struct.calcsize(fmt)
        result = struct.pack('II'+fmt, dtype, dataSize, *args)
        self._write(result.ljust(self.next8(len(result)),'\0'))
        return self.next8(dataSize + 8)

    
    def packStr(self, s, maxlen=31):
        """ Write a string to the file, proceeded by type and size info, aligned
            to 64 bits. Used internally.
        """
        s = s[:maxlen].encode('utf8')
        n = self.next8(len(s))
        fmt = 'II %ds' % n
        self._write(struct.pack(fmt, MP.miINT8, len(s), s))

   
    def endArray(self):
        """ End an array, updating all the sizes.
        """
        if not self._inArray:
            return False
        self._inArray = False
        
        dataEndPos = self.stream.tell()
        arrayEndPos = self.next8(dataEndPos)
        
        # Move back and rewrite the actual total size
        self._seek(self.dataStartPos-4)
        self._write(self.intPack.pack(self.next8(dataEndPos-self.dataStartPos)))
        
        # Move back and rewrite the actual number of rows (columns, actually)
        self._seek(self.rowsPos)
        self._write(self.intPack.pack(self.numRows))
            
        # Move back and write the actual payload size (the 'real' portion only)
        self._seek(self.prSize)
        self._write(self.intPack.pack(self.numRows * self.rowFormatter.size))
        
        # Go back to the end.
        self._seek(dataEndPos)
        if dataEndPos < arrayEndPos:
            self._write('\0' * (arrayEndPos-dataEndPos))

        if self.arrayHasTimes and self.arrayNoTimes and self.arrayStartTime is not None:
            self.writeValue('%s_start' % self.arrayBaseName, self.arrayStartTime, MP.miUINT64)
            
        self.arrayMType = None
        self.arrayColNames = None
        self.arrayStartTime = None

        return True


    def startArray(self, name, cols, rows=1, arrayNumber=0, 
                   mtype=None, dtype=MP.miDOUBLE, flags=0, noTimes=True,
                   hasTimes=True, colNames=None):
        """ Begin a 2D array for storing the recorded data.
        
            :param name: The name of the matrix (array).
            :param cols: The number of columns in the data (excluding time).
            :keyword rows: The number of rows in the data (if known).
            :keyword dtype: The Matlab data type in the matrix.
            :keyword mtype: The Matlab matrix type. Defaults to match `dtype`.
            :keyword flags: A set of bit flags for the matrix.
        """
        if self._inArray:
            self.endArray()

        self.arrayBaseName = sanitizeName(name)
        
        # Calculate size of data, plus any applicable 'header' data that needs
        # to be in the same file as the array (column names, start time offset, 
        # etc.)
        headerSize = 0
        if colNames is not None:
            headerSize += self.getNamesSize(colNames, '%s_names' % name, noTimes)
        if noTimes and hasTimes:
            headerSize += self.getValueSize('%s_start' % self.arrayBaseName, self.arrayStartTime, MP.miUINT64)
            
        newSize = 80 + self.next8(len(name)) + headerSize
        self.checkFileSize(newSize)

        self._inArray = True
        self.arrayNumber = arrayNumber
        self.numRows = 0
        self.expectedRows = rows
        self.arrayMType = mtype
        self.arrayDType = dtype
        self.arrayFlags = flags
        self.arrayNoTimes = noTimes
        self.arrayHasTimes = hasTimes
        self.arrayColNames = colNames
        self.arrayStartTime = None
        
        if arrayNumber > 0:
            name = "%s%d" % (self.arrayBaseName, arrayNumber)
            
        if self.arrayColNames:
            self.writeNames(self.arrayColNames, '%s_names' % self.arrayBaseName, self.arrayNoTimes)
        
        if self.arrayMType is None:
            self.arrayMType = self.classTypes[self.arrayDType]
        

        if self.arrayNoTimes:
            self.numCols = cols
        else:
            self.numCols = cols+1
        fchar = self.typeFormatChars.get(self.arrayDType, self.typeFormatChars[MP.miDOUBLE]) * self.numCols
        self.rowFormatter = struct.Struct(fchar)
        
        # Start of matrix element, initial size of 0 (rewritten at end)
        self._write(struct.pack("II", MP.miMATRIX, 0)) # Start
        self.dataStartPos = self.stream.tell()
        
        # Write flags and matrix type
        self.pack('BBBBBBBB', (self.arrayMType, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00))
        
        # Write matrix dimensions. Because the file stores data column first,re
        # the recording data is stored 'sideways': lots of columns. The
        # second dimension is rewritten at the end.
        self.pack('ii', (self.numCols, rows), dtype=MP.miINT32)
        self.rowsPos = self.stream.tell() - 4
        
        # Write the matrix name
        self.packStr(sanitizeName(name))
        
        # Write the start of the 'PR' element; the size will be filled in later.
        self._write(struct.pack('II', self.arrayDType, self.rowFormatter.size * rows))
        self.prSize = self.stream.tell() - 4
        

    def getStringArraySize(self, title, strings):
        """ Get the size of an array of strings before writing it.
        """
        textSize = max([len(s) for s in strings])
        return 40 + self.next8(textSize*len(strings)) + 8 + self.next8(len(title))
        
        
    def writeStringArray(self, title, strings):
        """ Write a set of strings as a MATLAB character array.
        """
        textSize = max([len(s) for s in strings])
        strings = [n.ljust(textSize) for n in strings]
        payload = ''.join([''.join(x) for x in zip(*strings)])
        
        totalSize = self.getStringArraySize(title, strings)
        
        # Ensure that this won't exceed the max file size before writing
        self.checkFileSize(totalSize)
        
        self._write(struct.pack("II", MP.miMATRIX, totalSize)) # Start
        self.pack('BBBBBBBB', (0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00))
        self.pack('II', (len(strings), textSize), dtype=MP.miINT32)
        self.packStr(sanitizeName(title))
        self._write(struct.pack("II", MP.miUTF8, textSize*len(strings)))
        self._write(payload.ljust(self.next8(len(payload)), '\0'))
        
    
    def getNamesSize(self, names, title="channel_names", noTimes=False):
        if noTimes:
            return self.getStringArraySize(title, names)
        return self.getStringArraySize(title, ['Time']+names)
    
        
    def writeNames(self, names, title="channel_names", noTimes=False):
        """ Write IDE column names to the MAT file, for easy identification
            of rows in MATLAB (IDE data is written in columns in order to
            stream).
        """
        if not noTimes:
            names = names[:]
            names.insert(0, 'Time')
        self.writeStringArray(title, names)


    def writeRow(self, event):
        """ Write a sample to the array.
        
            :param event: The sample, in the format `(time, (v1, v2, ...))`
        """
        if self.arrayStartTime is None:
            self.arrayStartTime = event[0]
        self.checkFileSize(self.rowFormatter.size, pad=96)

        try:
            if self.arrayNoTimes:
                data = self.rowFormatter.pack(*event[:1])
            else:
                data = self.rowFormatter.pack(event[0]*self.timeScalar, *event[1:])
        except struct.error as err:
            logger.exception("ERROR: %s, formatter=%r, event=%r" % (self.arrayBaseName, self.rowFormatter.format, event))
            raise err
            
        self._write(data)
        self.numRows += 1


    def getValueSize(self, name, val, dtype=MP.miDOUBLE):
        """ Get the total size of a value as written.
        """
        name = sanitizeName(name)
        fchar = self.typeFormatChars.get(dtype,'d')
        dsize = struct.calcsize(fchar)
        return 40 + self.next8(len(name)) + 8 + self.next8(dsize)
        

    def writeValue(self, name, val, dtype=MP.miDOUBLE):
        """ Write a single numeric value to the MAT file. Don't use for strings.
        """
        name = sanitizeName(name)
        mtype = self.classTypes.get(dtype, MP.mxDOUBLE_CLASS)
        fchar = self.typeFormatChars.get(dtype,'d')
        totalSize = self.getValueSize(name, val, dtype)
        
        self.checkFileSize(totalSize)
        
        self._write(struct.pack("II", MP.miMATRIX, totalSize)) # Start
        self.pack('BBBBBBBB', (mtype, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00)) #flags
        self.pack('II', (1, 1), dtype=MP.miINT32) # dimensions
        self.packStr(name)
        self.pack(fchar, (val,), dtype=dtype)

    
    def writeRecorderInfo(self, info):
        """ Write an IDE file's 'RecorderInfo' data to the MAT as a set of
            strings.
        """
        if info:
            s = ['%s:\t%s' % i for i in info.items()]
            self.writeStringArray('recorder_info', s)


    def _writeCalData(self, name, cal, doc=None):
        if isinstance(cal, int):
            cal = doc.transforms.get(cal, None)
        if cal is None:
            return
        vals = cal.references + cal.coefficients
        self.startArray(name, len(vals), dtype=MP.miDOUBLE, noTimes=True, hasTimes=False)
        self.writeRow((0,vals))
        self.endArray()


    def writeCalibration(self, cals):
        """ Write a dictionary of Transform objects (keyed by ID) to the file.
        """
#         print "writeCalibration %s" % self.stream.name
        for calId, cal in cals.items():
            self._writeCalData("calibration%d" % calId, cal)
    
    
    def writeCalPerChannel(self, doc):
        """ Write the calibration used by a document's Channels/Subchannels.
        """
        channels = [c for c in doc.channels.values() if self.calChannels is None or c.id in self.calChannels]
        for c in channels:
            self._writeCalData("%s_calibration" % c.displayName, c.transform, doc)
            for subc in c.subchannels:
                self._writeCalData("%s_calibration" % subc.displayName, subc.transform, doc)

    
    def close(self):
        """ Close the file.
        """
        if self._inArray:
            self.endArray()
        return self.stream.close()



#===============================================================================
# 
#===============================================================================

    @classmethod
    def makeHeader(cls, doc, session=-1, prefix="MATLAB 5.0 MAT-file"):
        """ Generate MAT file header text from a `Dataset` document.
        """
        if not isinstance(prefix, (str, bytes, bytearray)):
            prefix = ''
        elif not prefix.endswith(' '):
            prefix += ' '
            
        msg = "%sGenerated from %s" % (prefix, os.path.basename(doc.filename))
    
        s = doc.sessions[session]
        if s.utcStartTime:
            createTime = s.utcStartTime
        else:
            try:
                createTime = os.path.getctime(doc.filename)
            except IOError:
                createTime = None
        
        if createTime is not None:
            msg = "%s recorded %s UTC" % (msg, datetime.utcfromtimestamp(createTime))
        
        return msg


#===============================================================================
# 
#===============================================================================

def exportMat(events, filename, start=0, stop=-1, step=1, subchannels=True,
              callback=None, callbackInterval=0.01, timeScalar=1,
              raiseExceptions=False, useUtcTime=False, headers=True, 
              removeMean=None, meanSpan=None, display=False, matArgs={},
              noBivariates=False, **kwargs):
    """ Export a `dataset.EventList` as a Matlab .MAT file. Works in a manner
        similar to the standard `EventList.exportCsv()` method.
    
        :param events: an `EventList` from which to export.
        :param filename: The path/name of the .MAT file to write.
        :keyword start: The first event index to export (defaults to first).
        :keyword stop: The last event index to export (defaults to last).
        :keyword step: The number of events between exported lines.
        :keyword subchannels: A sequence of individual subchannel numbers
            to export. Only applicable to objects with subchannels.
            `True` (default) exports them all.
        :keyword callback: A function (or function-like object) to notify
            as work is done. It should take four keyword arguments:
            `count` (the current line number), `total` (the total number
            of lines), `error` (an exception, if raised during the
            export), and `done` (will be `True` when the export is
            complete). If the callback object has a `cancelled`
            attribute that is `True`, the MAT export will be aborted.
            The default callback is `None` (nothing will be notified).
        :keyword callbackInterval: The frequency of update, as a
            normalized percent of the total lines to export.
        :keyword timeScalar: A scaling factor for the even times.
            The default is 1 (microseconds).
        :keyword raiseExceptions: If `False`, all exceptions will be handled
            quietly, passed along to the callback.
        :keyword useUtcTime: If `True`, times are written as the UTC
            timestamp. If `False`, times are relative to the recording.
        :keyword removeMean: If `True`, remove the mean from the output.
        :keyword meanSpan: The span over which the mean is calculated. -1
            for the total mean.
        :keyword display: If `True`, export using the EventList's 'display'
            transform (e.g. unit conversion).
        :keyword matArgs: A dictionary of keyword arguments supplied to the
            `MatStream` constructor.
        :return: Tuple: The number of rows exported and the elapsed time.
    """
    noCallback = callback is None
    events = events.copy()
    
    events.noBivariates = noBivariates
    if removeMean is not None:
        events.removeMean = removeMean
    if meanSpan is not None:
        events.rollingMeanSpan = meanSpan
    
    start = (1 + start + len(events)) if start < 0 else start
    stop = (1 + stop + len(events)) if stop < 0 else stop
    totalLines = (stop - start) / (step + 0.0)
    updateInt = int(totalLines * callbackInterval)
    
    # Catch all or no exceptions
    ex = None if raiseExceptions or noCallback else Exception
    
    t0 = datetime.now()
    
    createTime = 0
    if useUtcTime:
        if events.session.utcStartTime:
            createTime = events.session.utcStartTime
        else:
            try:
                createTime = os.path.getctime(events.dataset.filename)
            except IOError:
                pass
            
        # Scale to increments used in the source.
        createTime /= timeScalar
   
    # If specific subchannels are specified, export them in order.
    if events.hasSubchannels:
        if subchannels is True:
            numCols = len(events.parent.subchannels)
            formatter = None
            names = [x.name for x in events.parent.subchannels]
        else:
            numCols = len(subchannels)
            # Create a function instead of chewing the subchannels every time
            formatter = eval("lambda x: (%s,)" % \
                       ",".join([("x[%d]" % c) for c in subchannels]))
            names = [events.parent.subchannels[x].name for x in subchannels]
    else:
        numCols = 1
        formatter = lambda x: (x,)
        names = [events.parent.name]

    totalSamples = totalLines * numCols
    if headers is False:
        names = None
    
    comments = MatStream.makeHeader(events.dataset, events.session.sessionId)
    matfile = MatStream(filename, events.dataset, comments, 
                        timeScalar=timeScalar, **matArgs)
    
    matfile.startArray(events.parent.name, numCols, rows=totalLines, 
                       colNames=names, noTimes=False)
    
    try:
        for num, evt in enumerate(events.iterSlice(start, stop, step, display)):
            t, v = evt[0], evt[1:]
            if formatter is not None:
                v = formatter(v)

            matfile.writeRow((createTime + t,)+v)
            
            if callback is not None:
                if getattr(callback, 'cancelled', False):
                    callback(done=True)
                    break
                if updateInt == 0 or num % updateInt == 0:
                    callback(num*numCols, total=totalSamples, 
                             filename=matfile.filename)
                    
        if callback:
            callback(done=True)
            
    except ex as e:
        callback(error=e)
        
    matfile.close()
    
    return num+1, datetime.now() - t0

#===============================================================================
# 
#===============================================================================




"""

"""

from datetime import datetime
import os.path
import sys
from time import time as time_time
from time import sleep

import struct

from . import transforms
from .dataset import Dataset
from . import parsers

#===============================================================================
# 
#===============================================================================

# from dataset import __DEBUG__

import logging
logger = logging.getLogger('idelib-archive')
logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s")


#===============================================================================
# Defaults
#===============================================================================

# from parsers import AccelerometerParser

# Legacy hard-coded sensor/channel mapping. Used when importing files recorded 
# on SSX running old firmware, which does not contain self-description data.
DEFAULTS = {
    "sensors": {
        0x00: {"name": "832M1 Accelerometer"},
        0x01: {"name": "MPL3115 Temperature/Pressure"}
    },
    
    "channels": {
            0x00: {"name": "Accelerometer XYZ",
    #                 "parser": struct.Struct("<HHH"), 
    #                 "transform": 0, #calibration.AccelTransform(),
                    "parser": struct.Struct("<HHH"),
                    "transform": 0,
                    "subchannels":{0: {"name": "Accelerometer Z", 
                                       "axisName": "Z",
                                       "units":('Acceleration','g'),
                                       "displayRange": (-100.0,100.0),
                                       "transform": 3,
                                       "warningId": [0],
                                       "sensorId": 0,
                                     },
                                  1: {"name": "Accelerometer Y", 
                                      "axisName": "Y",
                                      "units":('Acceleration','g'),
                                      "displayRange": (-100.0,100.0),
                                      "transform": 2,
                                      "warningId": [0],
                                      "sensorId": 0,
                                      },
                                  2: {"name": "Accelerometer X", 
                                      "axisName": "X",
                                      "units":('Acceleration','g'),
                                      "displayRange": (-100.0,100.0),
                                      "transform": 1,
                                      "warningId": [0],
                                      "sensorId": 0,
                                      },
                                },
                   },
            0x01: {"name": "Pressure/Temperature",
                   "parser": parsers.MPL3115PressureTempParser(),
                   "subchannels": {0: {"name": "Pressure", 
                                       "units":('Pressure','Pa'),
                                       "displayRange": (0.0,120000.0),
                                      "sensorId": 1,
                                       },
                                   1: {"name": "Temperature", 
                                       "units":('Temperature','\xb0C'),
                                       "displayRange": (-40.0,80.0),
                                      "sensorId": 1,
                                       }
                                   },
                   "cache": True,
                   "singleSample": True,
                   },
    },
    
    "warnings": [{"warningId": 0,
                   "channelId": 1,
                   "subchannelId": 1,
                   "low": -20.0,
                   "high": 60.0
                   }]
}


if False: #__DEBUG__:
    logger.info("Adding low g channels to defaults")
    DEFAULTS['sensors'][0x02] = {
         'name': "Low-G Accelerometer"
    }
    DEFAULTS['channels'].update({
        0x02: {'name': "Low-G Accelerometer XYZ",
               'parser': struct.Struct(">hhh"),
               "subchannels":{0: {"name": "Low-g Z", 
                                  "units":('Acceleration','g'),
                                  "sensorId": 2,
                                 },
                              1: {"name": "Low-g Y", 
                                  "units":('Acceleration','g'),
                                  "sensorId": 2,
                                  },
                              2: {"name": "Low-g X", 
                                  "units":('Acceleration','g'),
                                  "sensorId": 2,
                                  },
                            },
               },
#         0x43: {"name": "DEBUG Crystal Drift",
#                "parser": struct.Struct(">II")},
#         0x45: {"name": "DEBUG Gain/Offset",
#                "parser": struct.Struct("<i")},
    })


def createDefaultSensors(doc, defaults=DEFAULTS):
    """ Given a nested set of dictionaries containing the definition of one or
        more sensors, instantiate those sensors and add them to the dataset
        document.
    """
#     logger.info("creating default sensors")
    sensors = defaults['sensors'].copy()
    channels = defaults['channels'].copy()
    warnings = defaults['warnings']
    
    if doc.recorderInfo:
        # TODO: Move device-specific stuff out of the main importer
        rtype = doc.recorderInfo.get('RecorderTypeUID', 0x10)
        if rtype | 0xff == 0xff:
            # SSX recorders have UIDs that are zero except the least byte.
            SSX_ACCEL_RANGES = {
               0x10: (-25,25),
               0x12: (-100,100),
               0x13: (-200,200),
               0x14: (-500, 500),
               0x15: (-2000, 2000),
               0x16: (-6000, 6000)
            }
            rrange = SSX_ACCEL_RANGES.get(rtype & 0xff, (-25,25))
            transform = transforms.AccelTransform(*rrange)
            ch0 = channels[0x00]
            ch0['transform'] = transform
            for i in range(3):
                ch0['subchannels'][i]['displayRange'] = rrange

    for sensorId, sensorInfo in sensors.items():
        doc.addSensor(sensorId, sensorInfo.get("name", None))
        
    for chId, chInfo in channels.items():
        chArgs = chInfo.copy()
#         chArgs['sensor'] = sensor
        subchannels = chArgs.pop('subchannels', None)
        channel = doc.addChannel(chId, **chArgs)
        if subchannels is None:
            continue
        for subChId, subChInfo in subchannels.items():
            channel.addSubChannel(subChId, **subChInfo)
    
    for warn in warnings:
        doc.addWarning(**warn)
    

#===============================================================================
# Parsers/Element Handlers
#===============================================================================

# Parser importer. These are taken from the module by type. We may want to 
# create the list of parser types 'manually' prior to release; it's marginally 
# safer.
elementParserTypes = parsers.getElementHandlers()


def instantiateParsers(doc, parserTypes=elementParserTypes):
    """ Create a dictionary of element parser objects keyed by the name of the
        element they handle. Handlers that handle multiple elements have
        individual keys for each element name.
    """
    elementParsers = {}
    for t in parserTypes:
        p = t(doc)
        if isinstance(t.elementName, str):
            elementParsers[t.elementName] = p
        else:
            for name in t.elementName:
                elementParsers[name] = p
    return elementParsers


#===============================================================================
# Updater callbacks
#===============================================================================

def nullUpdater(*args, **kwargs):
    """ A progress updater stand-in that does nothing. """
    if kwargs.get('error',None) is not None:
        raise kwargs['error']
nullUpdater.cancelled = False
nullUpdater.paused = False


class SimpleUpdater(object):
    """ A simple text-based progress updater.
        :ivar cancelled: If set to `True`, the job using the updater will abort. 
        :ivar paused: If set to `True`, the job using the updater will pause.
    """
    
    def __init__(self, cancelAt=1.0, quiet=False):
        """ Constructor.
            :keyword cancelAt: A percentage at which to abort the import. For
                testing purposes.
        """
        self.cancelled = False
        self.paused = False
        self.startTime = None
        self.cancelAt = cancelAt
        self.estSum = None
        self.quiet = quiet
    
    def dump(self, s):
        if not self.quiet:
            sys.stdout.write(s)
    
    def __call__(self, count=0, total=None, percent=None, error=None, 
                 starting=False, done=False):
        if percent >= self.cancelAt:
            self.cancelled=True
        if self.startTime is None:
            self.startTime = datetime.now()
        if starting:
            logger.info("Import started at %s" % self.startTime)
            return
        if done:
            logger.info("Import completed in %s" % (datetime.now() - self.startTime))
            logger.info("Original estimate was %s" % self.estSum)
        else:
            self.dump('\x0d%s samples read' % count)
            if percent is not None:
                p = int(percent*100)
                self.dump(' (%d%%)' % p)
                if p > 0 and p < 100:
                    d = ((datetime.now() - self.startTime) / p) * (100-p)
                    self.dump(' - est. completion in %s' % d)
                    if self.estSum is None:
                        self.estSum = d
                else:
                    self.dump(' '*25)
            sys.stdout.flush()
    

#===============================================================================
# ACTUAL FILE READING HAPPENS BELOW
#===============================================================================

def importFile(filename='', updater=nullUpdater, numUpdates=500,
               updateInterval=1.0, parserTypes=elementParserTypes, 
               defaults=DEFAULTS, name=None, quiet=False):
    """ Create a new Dataset object and import the data from a MIDE file. 
        Primarily for testing purposes. The GUI does the file creation and 
        data loading in two discrete steps, as it will need a reference to 
        the new document before the loading starts.
        :see: `readData()`
    """
    stream = open(filename, "rb")
    doc = openFile(stream, updater=updater, name=name, parserTypes=parserTypes,
                   defaults=defaults, quiet=quiet)
    readData(doc, updater=updater, numUpdates=numUpdates, 
             updateInterval=updateInterval, parserTypes=parserTypes, 
             defaults=defaults)
    return doc


def openFile(stream, updater=nullUpdater, parserTypes=elementParserTypes,  
             defaults=DEFAULTS, name=None, quiet=False, getExitCond=True):
    """ Create a `Dataset` instance and read the header data (i.e. non-sample-
        data). When called by a GUI, this function should be considered 'modal,' 
        in that it shouldn't run in a background thread, unlike `readData()`. 
        
        :param stream: The file or file-like object containing the EBML data.
        :keyword updater: A function (or function-like object) to notify as 
            work is done. It should take four keyword arguments: `count` (the 
            current line number), `total` (the total number of samples), `error` 
            (an unexpected exception, if raised during the import), and `done` 
            (will be `True` when the export is complete). If the updater object 
            has a `cancelled` attribute that is `True`, the import will be 
            aborted. The default callback is `None` (nothing will be notified).
        :keyword parserTypes: A collection of `parsers.ElementHandler` classes.
        :keyword defaults: A nested dictionary containing a default set 
            of sensors, channels, and subchannels. These will only be used if
            the dataset contains no sensor/channel/subchannel definitions. 
        :keyword name: An optional name for the Dataset. Defaults to the
            base name of the file (if applicable).
        :keyword quiet: If `True`, non-fatal errors (e.g. schema/file
            version mismatches) are suppressed. 
        :return: The opened (but still 'empty') `dataset.Dataset`
    """
    if isinstance(stream, str):
        stream = open(stream, 'rb')
    
    if getExitCond:
        exitCond = getExitCondition(stream)
    else:
        exitCond = None
   
    doc = Dataset(stream, name=name, exitCondition=exitCond, quiet=quiet)
    doc.addSession()

    if doc._parsers is None:
        doc._parsers = instantiateParsers(doc, parserTypes)
    
    elementParsers = doc._parsers
    
    try:
        for r in doc.ebmldoc:
            if getattr(updater, "cancelled", False):
                doc.loadCancelled = True
                break
            if r.name not in elementParsers:
                continue
            parser = elementParsers[r.name]
            if parser.makesData():
                break
            parser.parse(r) 
            
    except IOError as e:
        if e.errno is None:
            # The EBML library raises an empty IOError if it hits EOF.
            # TODO: Handle other cases of empty IOError (lots in python-ebml)
            doc.fileDamaged = True
        else:
            updater(error=e)

    except TypeError as e:
        logger.exception(e)
        # This can occur if there is a bad element in the data
        # (typically the last)
        doc.fileDamaged = True

    if not doc.sensors:
        # Got data before the recording props; use defaults.
        if defaults is not None:
            createDefaultSensors(doc, defaults)
            
    doc.updateTransforms()
    return doc


def readData(doc, source=None, updater=nullUpdater, numUpdates=500, updateInterval=.1,
             total=None, bytesRead=0, samplesRead=0, parserTypes=elementParserTypes,
             sessionId=0, onlyChannel=None, maxPause=None, **kwargs):
    """ Import the data from a file into a Dataset.
    
        :param doc: The Dataset document into which to import the data.
        :param source: An alternate Dataset to merge into the main one.
        :keyword updater: A function (or function-like object) to notify as 
            work is done. It should take four keyword arguments: `count` (the 
            current line number), `total` (the total number of samples), `error` 
            (an unexpected exception, if raised during the import), and `done` 
            (will be `True` when the export is complete). If the updater object 
            has a `cancelled` attribute that is `True`, the import will be 
            aborted. The default callback is `None` (nothing will be notified).
        :keyword numUpdates: The minimum number of calls to the updater to be
            made. More updates will be made if the updates take longer than
            than the specified `updateInterval`. 
        :keyword updateInterval: The maximum number of seconds between calls to 
            the updater. More updates will be made if indicated by the specified
            `numUpdates`.
        :keyword total: The total number of bytes in the file(s) being imported.
            Defaults to the size of the current file, but can be used to
            display an overall progress when merging multiple recordings. For
            display purposes.
        :keyword bytesRead: The number of bytes already imported. Mainly for
            merging multiple recordings. For display purposes.
        :keyword samplesRead: The total number of samples imported. Mainly for
            merging multiple recordings.
        :keyword parserTypes: A collection of `parsers.ElementHandler` classes.
        :keyword sessionId: The Session number to import. For future use; it
            currently does nothing, and SSX files currently do not contain
            multiple sessions.
        :keyword onlyChannel: If supplied, only import data from one channel.
        :keyword maxPause: If the updater's `paused` attribute is `True`, the
            import will pause. This is the maximum pause length.
        :return: The total number of samples read.
    """
    
    if doc._parsers is None:
        doc._parsers = instantiateParsers(doc, parserTypes)
    
    elementParsers = doc._parsers

    elementCount = 0
    eventsRead = 0
    
    # Progress display stuff
    if total is None:
        total = doc.ebmldoc.size + bytesRead
        
    dataSize = total
    
    if numUpdates > 0:
        ticSize = total / numUpdates 
    else:
        # An unreachable file position effectively disables the updates.
        ticSize = total+1
    
    if updateInterval > 0:
        nextUpdateTime = time_time() + updateInterval
    else:
        # An update time 24 hours in the future should mean no updates.
        nextUpdateTime = time_time() + 5184000
    
    firstDataPos = 0
    nextUpdatePos = bytesRead + ticSize
    
    timeOffset = 0
    maxPause = getattr(updater, "maxPause", maxPause)
    
    # Actual importing ---------------------------------------------------------
    if source is None:
        source = doc
    try:    
        # This just skips 'header' elements. It could be more efficient, but
        # the size of the header isn't significantly large; savings are minimal.
        for r in source.ebmldoc:
            
            r_name = r.name
            if r_name == "ChannelDataBlock":
                r_name = "ChannelDataArrayBlock"
            
            doc.loadCancelled = getattr(updater, "cancelled", False)
            if doc.loadCancelled:
                break
            
            if updater.paused:
                # Pause or throttle import.
                pauseTime = time_time()
                while updater.paused:
                    sleep(0.125)
                    if maxPause and time_time() - pauseTime > maxPause:
                        break
            
            if r_name not in elementParsers:
                # Unknown block type, but probably okay.
                logger.info("unknown block %r (ID 0x%02x) @%d" % \
                            (r_name, r.id, r.offset))
                continue
            
            # HACK: Not the best implementation. Should be moved somewhere.
            if onlyChannel is not None and (r_name == "ChannelDataArrayBlock"):
                if r.value[0].value != onlyChannel:
                    continue 
            
            if source != doc and r_name == "TimeBaseUTC":
                timeOffset = (r.value - doc.lastSession.utcStartTime) * 1000000.0
                continue
                
            try:
                parser = elementParsers[r_name]
                if not parser.isHeader or r_name == "Attribute":
                    added = parser.parse(r, timeOffset=timeOffset)
                    if isinstance(added, int):
                        eventsRead += added
                    
            except parsers.ParsingError as err:
                # TODO: Error messages
                logger.error("Parsing error during import: %s" % err)
                continue

            elementCount += 1
            
            # More progress display stuff -------------------------------------
            # FUTURE: Possibly do the update check every nth elements; that
            # would have slightly less work per cycle.
            thisOffset = r.offset + bytesRead
            thisTime = time_time()
            if thisTime > nextUpdateTime or thisOffset > nextUpdatePos:
                # Update progress bar
                updater(count=eventsRead+samplesRead,
                        percent=(thisOffset-firstDataPos+0.0)/dataSize)
                nextUpdatePos = thisOffset + ticSize
                nextUpdateTime = thisTime + updateInterval
            
    except IOError as e:
        if e.errno is None:
            # The EBML library raises an empty IOError if it hits EOF.
            # TODO: Handle other cases of empty IOError (lots in python-ebml)
            doc.fileDamaged = True
        else:
            updater(error=e, done=True)
        
    except TypeError:
        # This can occur if there is a bad element in the data
        # (typically the last)
        doc.fileDamaged = True

    doc.loading = False
    updater(done=True)
    return eventsRead


#===============================================================================
# 
#===============================================================================

def estimateLength(filename, numSamples=50000, parserTypes=elementParserTypes, 
                   defaults=DEFAULTS):
    """ Open and read enough of a file to get a rough estimate of its complete
        time range. 
        
        :param filename: The IDE file to open.
        :keyword numSamples: The number of samples to read before generating
            an estimated size.
        :keyword parserTypes: A collection of `parsers.ElementHandler` classes.
        :keyword defaults: Default Slam Stick description data, for use with
            old SSX recordings.
        :return: A 3-element tuple: 
    """
    
    # Fake updater that just quits after some number of samples.
    class DummyUpdater(object):
        cancelled = False
        paused = False
        def __init__(self, n):
            self.numSamples = n
        def __call__(self, count=0, **kwargs):
            self.cancelled = count > self.numSamples
                
    updater = DummyUpdater(numSamples)
    
    with open(filename, "rb") as stream:
        doc = openFile(stream, parserTypes=parserTypes,
                       defaults=defaults, quiet=True)
        dataStart = stream.tell()
        totalSize = os.path.getsize(filename) - dataStart
        
        # read a portion of the recording
        readData(doc, updater=updater, parserTypes=parserTypes, 
                 defaults=defaults)
        chunkSize = stream.tell() - dataStart
        
        start = sys.maxsize
        end = -1
        numEvents = 0.0
        for ch in doc.channels.values():
            events = ch.getSession()
            if len(events) > 0:
                start = min(start, events[0][0])
                end = max(end, events[-1][0])
                numEvents += (len(events) * len(ch.subchannels))
        
        chunkTime = end - start
        
    return start, start + (totalSize / chunkSize * chunkTime), numEvents/chunkTime
    

#===============================================================================
# 
#===============================================================================

    
def getExitCondition(recording, bytesRead=1000):
    """ Get the ``ExitCond`` Attribute from the end of a recording, if present.
        The result will be an integer:
        
        * 1: Button press
        * 2: USB connection
        * 3: Recording time limit reached
        * 4: Low battery
        * 5: File size limit reached
        * 128: I/O error (can occur if disk is full or 4GB FAT32 size limit
          reached.
        
        :param recording: The IDE file, either a filename or a file-like object.
        :keyword bytesRead: The number of bytes to read from the end of the
            recording file.
    """
    result = None
    
    if isinstance(recording, str):
        with open(recording, "rb") as fs:
            return getExitCondition(fs)

    filename = recording.name
    offset = recording.tell()
   
    recording.seek(os.path.getsize(filename) - bytesRead)
    data = recording.read()
    try:
        idx = data.index(b"ExitCond") + 11
        if idx <= len(data):
            result = data[idx]
    except (IOError, IndexError, ValueError) as e:
        logger.warning(e)

    recording.seek(offset)
    return result        



"""
This file contains three types of things: special-case data parsers, handlers
for specific EBML elements, and wrapper classes for EBML data-containing
elements.

Special-case data-parsing objects are for parsing element data needing more 
post-processing than simply grabbing values with a `struct.Struct`, such as a 
sensor that produces data in a non-power-of-two length (e.g. 24 bits). 
Special-case parsers must (to a limited degree) quack like `struct.Struct` 
objects; they must provide `size` and `format` attributes (the latter should be 
`None`), and they must implement the method `unpack_from()`. They may also 
include optional `types` and `ranges` attributes, which are used to provide
'hints' about the data. These are automatically computed if not defined
explicitly.

Element handlers are called by the importer as it iterates through the 'root'
elements of an EBML file. Generally, handlers are instantiated only once, just
after the new `Dataset` has been created. Handlers of data-containing elements 
(e.g. `ChannelDataBlock` and `SimpleChannelDataBlock`) operate as factories, 
manufacturing instances of their respective wrapper objects. All handlers have 
an `elementName` attribute containing the name of the EBML element on which 
they work. A single handler class handles a single element type. Data-producing 
handlers also have a `product` attribute, which is a reference to the class of 
the object they manufacture. It is up to the handler to handle any children of 
their respective elements.

Data elements wrap EBML elements and abstract the details of getting at their 
data. A sensor that outputs the same data in both normal `ChannelDataBlock` 
and `SimpleChannelDataBlock` elements are represented the same way in the API.

See the `dataset` module for more information.

:todo: Clean this up! it's grown a little too organically.

Created on Sep 26, 2013
:author: dstokes
"""

from collections import OrderedDict
from collections.abc import Sequence
import math
import struct
import sys
import types

import numpy as np  

from . import transforms
from .util import decode_attributes

import logging
logger = logging.getLogger('idelib-archive')
logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s")

#===============================================================================
# 
#===============================================================================

DATA_PARSERS = {}

def dataParser(cls):
    """ Decorator. Used to register classes as parsers of data payloads. 
    """
    global DATA_PARSERS
    DATA_PARSERS[cls.__name__] = cls
    return cls


#===============================================================================
# Utility Functions
#===============================================================================            
    
def renameKeys(d, renamed, exclude=True, recurse=True, ordered=False,
               mergeAttributes=True):
    """ Create a new dictionary from and old one, using different keys. Used
        primarily for converting EBML element names to function keyword
        arguments.
    
        :param d: The source dictionary
        :param renamed: A dictionary of new names keyed by old names
        :keyword exclude: If `True`, only keys appearing in `renamed` are
            copied to the new dictionary.
        :keyword recurse: If `True`, the renaming operates over all nested
            dictionaries.
        :keyword ordered: If `True`, the results are returned as an
            `OrderedDict` rather than a standard dictionary.
        :keyword mergeAttributes: If `True`, any `Attribute` elements are
            processed into a standard key/values and merged into the main
            dictionary.
        :return: A new dictionary, a deep copy of the original, with different
            keys.
    """
    if isinstance(d, (str, bytes, bytearray)):
        return d
    elif isinstance(d, Sequence):
        return [renameKeys(i, renamed, exclude, recurse, ordered, mergeAttributes) for i in d]
    elif not isinstance(d, dict):
        return d
    
    if ordered:
        result = OrderedDict()
    else:
        result = {}
        
    for oldname,v in d.items():
        if oldname == "Attribute":
            if mergeAttributes:
                result.update(decode_attributes(v))
            else:
                result['Attribute'] = decode_attributes(v)
            continue
        
        if oldname not in renamed and exclude:
            continue
        
        newname = renamed.get(oldname, oldname)
        
        if recurse:
            result[newname] = renameKeys(v, renamed, exclude, recurse)
        else:
            result[newname] = v
            
    return result


def valEval(value, allowedChars='+-/*01234567890abcdefx.,() ;\t\n_'):
    """ A slightly safer version of `eval()` designed for simple mathematical
        expressions. Only permits math stuff.
    """
    # XXX: This is a hideous Rube Goldberg machine of a function. Could probably
    # be accomplished better with a suitably complex regular expression.
    if isinstance(value, (bytes, bytearray)):
        value = value.decode()
    value = value.strip()
    val = value.lower().replace('math.', '')
    
    # Allow functions from the math module. Replace them with a placeholder
    # character (128+). Do in reverse order of length to prevent partial
    # replacement (e.g. 'e' in 'exp', 'tan' in 'atan2'
    funcs = sorted([x for x in dir(math) if not x.startswith('_')],
                   key=lambda x: len(x), reverse=True)
    # funcs = tuple(enumerate(funcs, 128))
    
    for n, f in enumerate(funcs, 128):
        allowedChars = "%s%s" % (allowedChars, chr(n))
        if f in val:
            val = val.replace(f, chr(n))
            continue

    allowedChars = set(allowedChars)
        
    if len(set(val).difference(allowedChars)) > 0: 
        raise TypeError("valEval(): Invalid/Unsafe character in %r" % value)
    
    # Put the math functions back, with module reference
    if len(val) < len(value):
        for n, f in enumerate(funcs, 128):
            val = val.replace(chr(n), 'math.%s' % f)
    
    return eval(val)
    
    
def parseAttribute(obj, element, multiple=True):
    """ Utility function to parse an `Attribute` element's data into a 
        key/value pair and apply it to an object's `attribute` attribute
        (a dictionary).
        
        :param element: The `Attribute` element to parse.
        :keyword multiple: An object may have more than one Attribute element
            with the same name. If `True`, the value corresponding to the name
            is a list which is appended to. If `False`, the value is that of
            the last `Attribute` element parsed. 
    """
    if not hasattr(obj, 'attributes'):
        obj.attributes = OrderedDict()
        
    k = v = None
    for ch in element.value:
        if ch.name == "AttributeName":
            k = ch.value
        else:
            v = ch.value
            
    if k is not None:
        if multiple:
            obj.attributes.setdefault(k, []).append(v)
        else:
            obj.attributes[k] = v
        
    return k, v


#===============================================================================
# 
#===============================================================================

# The minimum and maximum values for data parsed out of data blocks.
# Used to provide an initial range, which is later corrected for actual
# values. Note that floats default to the normalized range of (-1.0, 1.0).
RANGES = {'c': None,
          'b': (-128,127),
          'B': (0,255),
          '?': (0,1),
          'h': (-(2**(8*2))/2, (2**(8*2))/2-1),
          'H': (0, 2**(8*2)-1),
          'i': (-(2**(8*4))/2, (2**(8*4))/2-1),
          'I': (0, 2**(8*4)-1),
          'l': (-(2**(8*4))/2, (2**(8*4))/2-1),
          'L': (0, 2**(8*4)-1),
          'q': (-(2**(8*8))/2, (2**(8*8))/2-1),
          'Q': (0, 2**(8*8)-1),
          'f': (-1.0, 1.0),
          'd': (-1.0, 1.0),
          'p': None,
          'P': None,
          's': None
}


def getParserTypes(parser):
    """ Get the Python types produced by a parser. If the parser doesn't
        explicitly define them in a `types` attribute, the types are
        derived from what the parser generates.
        
        :param parser: A `struct.Struct`-like parser.
        :return: A tuple of Python types.
    """
    if hasattr(parser, "types"):
        return parser.types
    return tuple([type(x) for x in parser.unpack_from(b'\xff'*parser.size)])


def getParserRanges(parser, useDefaults=True):
    """ Get the range of values created by a parser. If the parser doesn't
        explicitly define them in a `ranges` attribute, the theoretical 
        minimum and maximum values of the resulting data type are returned.
        Note that floating point values are typically reported as (-1.0,1.0).
        
        :param parser: A `struct.Struct`-like parser.
        :return: A tuple of (min, max) tuples. Non-numeric values will
            have a reported range of `None`.
    """
    if hasattr(parser, "ranges"):
        return parser.ranges
    if not useDefaults:
        return None
    ranges = []
    for c in parser.format:
        if isinstance(c, int):
            c = chr(c)
        if c in RANGES:
            ranges.append(RANGES[c])
    return tuple(ranges)
    

#===============================================================================
# 
#===============================================================================

def getElementHandlers(module=None, subElements=False):
    """ Retrieve all EBML element handlers (parsers) from a module. Handlers
        are identified by being subclasses of `ElementHandler`.
    
        :keyword module: The module from which to get the handlers. Defaults to
            the current module (i.e. `idelib-archive.parsers`).
        :keyword subElements: `True` if the set of handlers should also
            include non-root elements (e.g. the sub-elements of a
            `RecordingProperties` or `ChannelDataBlock`).
        :return: A list of element handler classes.
    """
    elementParserTypes = []
    if module is None:
        if __name__ in sys.modules:
            moduleDict = sys.modules[__name__].__dict__
        else:
            moduleDict = globals()
    else:
        moduleDict = module.__dict__
    for p in moduleDict.values():
        if not isinstance(p, type) or p == ElementHandler:
            continue
        if issubclass(p, ElementHandler):
            if subElements or not p.isSubElement:
                elementParserTypes.append(p)
    return elementParserTypes


#===============================================================================
# EXCEPTIONS
#===============================================================================

class ParsingError(IOError):
    """ Exception raised for any failure to parse an EBML file. """
    pass


#===============================================================================
#--- SENSOR DATA PARSING
#===============================================================================

@dataParser
class MPL3115PressureTempParser(object):
    """ A special-case channel parser for the native raw data generated by 
        the MPL3115 Pressure/Temperature sensor. See module docs for more
        information on custom parsers.

        Data format is 5 bytes in total:
            Pressure (3 bytes): 
                Bits [23..6] whole-number value (signed)
                Bits [5..4] fractional value (unsigned)
                Bits [3..0] (ignored)
            Temperature (2 bytes): 
                Bits [15..8] whole-number value (signed)
                Bits [7..4] fractional value (unsigned)
                Bits [3..0] (ignored)
        
        :todo: Make sure the unsigned fraction part is correct for negative 
            whole values. This currently assumes -1 and .5 is -0.5, not -1.5

        :deprecated: Used only for really old recordings without recorder
            description data. Later firmware writes the data padded to a more 
            easily handled form.
            
        :cvar size: The size (in bytes) of one parsed sample. Always `5`.
        :cvar format: For compatibility with `struct.Struct`. Always `None`
        :cvar ranges: A tuple containing the absolute min and max values.
        :cvar types: A tuple containing the types of data parsed.
    """

    # Custom parsers need to provide a subset of a struct.Struct's methods
    # and attributes: 
    size = 5
    format = None

    # The absolute min and max values, returned via `getParserRanges()`. This
    # must be implemented. Normal `struct.Struct` objects get this computed
    # from their formatting string.
    ranges = ((0.0,120000.0), (-40.0,80.0))
    
    # The types of data parsed from each channel, returned via 
    # `getParserTypes()`. If this doesn't exist, the types are computed. 
    types = (float, float)
    
    # This is weirdly formed data. Using two parsers over the same data is
    # cheaper than using one plus extra bit manipulation.
    _pressureParser = struct.Struct(">i")
    _parser = struct.Struct(">xxBbB")
        
    def unpack_from(self, data, offset=0):
        """ Special-case parsing of a temperature data block.
        """
        rawpressure = self._pressureParser.unpack_from(data, offset)[0] >> 14
        fracpressure, rawtemp, fractemp = self._parser.unpack_from(data, offset)
        fracpressure = ((fracpressure >> 4) & 0b11) * 0.25
        fractemp = (fractemp >> 4) * 0.0625

        return (rawpressure + fracpressure, rawtemp + fractemp)
            

@dataParser
class AccelerometerParser(object):
    """ Parser for the accelerometer data. Accelerometer values are recorded
        as uint16 but represent values -(max) to (max) G, with 'max' being
        a property of the specific accelerometer part number. This parser 
        performs the conversion on the fly.
        
        :deprecated: Used only for really old recordings without recorder
            description data.
        :cvar size: The size (in bytes) of one parsed sample.
        :cvar format: The `struct.Struct` parsing format string used to parse.
        :cvar ranges: A tuple containing the absolute min and max values.
    """

    def __init__(self, formatting="<HHH"):
        self.parser = struct.Struct(formatting)
        self.format = self.parser.format
        self.size = self.parser.size
        self.ranges = ((-32768, 32767),) * 3

    def unpack_from(self, data, offset=0):
        # This parser converts to signed ints to avoid some problems with the
        # inverted 'z' axis.
        z, y, x = self.parser.unpack_from(data, offset)
        return 32768-z, y-32768, x-32768
    

################################################################################
#===============================================================================
#--- Base element parsing and data storing classes
#===============================================================================
################################################################################

class ElementHandler(object):
    """ Base class for all element handlers (i.e. parsers). ElementHandlers are
        instantiated after a `Dataset` has been created, and their `parse()`
        methods add data to (or modify) that Dataset.
        
        :cvar product: The class of object generated by the handler (if any).
            Used by data block parsers.
        :cvar elementName: The name of the EBML element handled.
        :cvar isSubElement: `False` if the element occurs at the root level of
            the EBML file.
        :cvar children: A list/tuple of the parsers for sub-elements of this
            element, if any.
    """
    product = None
    elementName = None
    isSubElement = False
    isHeader = False
    children = None
    
    def __init__(self, doc, **kwargs):
        self.doc = doc
        
        # Initialize the sub-element handlers
        if self.children is not None:
            self.childHandlers = {}
            for parser in self.children:
                if parser.elementName is None:
                    continue
                p = parser(self.doc)
                if isinstance(parser.elementName, (str, bytes, bytearray)):
                    self.childHandlers[parser.elementName] = p
                else:
                    for n in parser.elementName:
                        self.childHandlers[n] = p
                
        
    def parse(self, element, **kwargs):
        """ Process an EBML element's contents. Handlers that generate data
            blocks should return the number of samples read. Other element
            types can return whatever their role requires.
        """ 
        # Apply the sub-element handlers (if any) to the element's payload
        if self.children is not None:
            result = []
            for el in element.value:
                if el.name in self.childHandlers:
                    handler = self.childHandlers[el.name]
                    result.append(handler.parse(el, **kwargs))
            return result

    
    def getElementName(self, element):
        """ Generate a string with an element's name, ID, and file position
            for debugging/error reporting.
        """
        try:
            return "%r (0x%x) @%r" % (element.name, element.id,
                                      element.offset)
        except AttributeError:
            # Possibly an old python-ebml element
            # TODO: Remove legacy code
            return "%r (0x%x) @%r" % (element.name, element.id,
                                      element.stream.offset)


    def makesData(self):
        """ Does this handler produce sample data?
        """
        return self.product is not None and \
            issubclass(self.product, BaseDataBlock)


################################################################################
#===============================================================================
#--- Data parsers and handlers
#===============================================================================
################################################################################

class BaseDataBlock(object):
    """ Base class for all data-containing elements. Created by the
        appropriate ElementHandler for the given EBML element type.
        
        :cvar maxTimestamp: The modulus of the data's timestamp
        :cvar timeScalar: The scaling factor to convert native units (i.e.
            clock ticks) to microseconds.
    """
    maxTimestamp = 2**24 
    timeScalar = 1000000.0 / 2**15
    
    def __init__(self, element, maxTimestamp=maxTimestamp, timeScalar=timeScalar):
        self.element = element
        
        # This stuff will vary based on parser:
        self.blockIndex = -1
        self.startTime = None
        self.endTime = None
        self.numSamples = None
        self.sampleRate = None
        self.sampleTime = None
        self.indexRange = None
#         self._len = None
        self.payloadSize = None
        self.cache = False
        
        self.minMeanMax = None
        self.min = None
        self.mean = None
        self.max = None
        self._rollingMean = None
        self._rollingMeanSpan = 5000000
        self._rollingMeanLen = None  # length of set at last rolling mean
        
        self.maxTimestamp = maxTimestamp
        self.timeScalar = timeScalar
        

    def __repr__(self):
        return "<%s Channel: %d>" % (self.__class__.__name__, self.getHeader()[1])


    def parseWith(self, parser, start=None, end=None, step=1, subchannel=None):
        """ Parse an element's payload. Use this instead of directly using
            `parser.parse()` for consistency's sake.
            
            :param parser: The DataParser to use
            :keyword start: First subsample index to parse 
            :keyword end: Last subsample index to parse
            :keyword step: The number of samples to skip, if the start and end
                cover more than one sample.
            :keyword subchannel: The subchannel to get, if specified.
        """
        # SimpleChannelDataBlock payloads contain header info; skip it.
        data = self.payload
        start, end, step = slice(start, end, step).indices(self.payloadSize // parser.size)

        start = (start*parser.size)
        end = (end*parser.size)
        
        parser_unpack_from = parser.unpack_from
        if subchannel is not None:
            for i in range(start,end,parser.size*step):
                yield parser_unpack_from(data, i)[subchannel]
        else:
            for i in range(start,end,parser.size*step):
                yield parser_unpack_from(data, i)


    def parseByIndexWith(self, parser, indices, subchannel=None):
        """ Parse an element's payload and get a specific set of samples. Used
            primarily for resampling tricks.
            
            :param parser: The DataParser to use
            :param indices: A list of indices into the block's data. 
            :keyword subchannel: The subchannel to get, if specified.
        """
        # SimpleChannelDataBlock payloads contain header info; skip it.
        data = self.payload
        parser_unpack_from = parser.unpack_from
        parser_size = parser.size
        for i in indices:
            if i >= self.numSamples:
                continue
            idx = (i*parser_size)
            if subchannel is not None:
                yield parser_unpack_from(data, idx)[subchannel]
            else:
                yield parser_unpack_from(data, idx)


    def parseMinMeanMax(self, parser):
        

        if self.minMeanMax is None:
            return None
        try:
            # NOTE: Because the SSX Z axis is inverted, the min/max need correction
            # This may create a performance hit. Optimize?
            self.min = parser.unpack_from(self.minMeanMax)
            self.mean = parser.unpack_from(self.minMeanMax, parser.size)
            self.max = parser.unpack_from(self.minMeanMax, parser.size*2)

            self.min, self.max = list(zip(
                *(sorted(minmax) for minmax in zip(self.min, self.max))
            ))

            return self.min, self.mean, self.max
        except struct.error:
            # Bad min/mean/max data: too short. Ignore it.
            self.minMeanMax = None
            return None


    def getNumSamples(self, parser):
        """ Compute the number of subsamples in an element's payload, as
            parsed by the given data parser.
        """
        if self.numSamples is not None:
            return self.numSamples
        self.numSamples = int(self.payloadSize/parser.size)
        return self.numSamples


    def isValidLength(self, parser):
        """ Check if an element's payload data is evenly divisible by into
            a set of channels.
        
            :param n: The size of the data in bytes.
            :return: `True` if the number of bytes can be evenly
                distributed into subsamples.
        """
        return self.payloadSize > 0 and self.payloadSize % parser.size == 0

    
    @property
    def payload(self):
        return self.element.value


#===============================================================================
# SimpleChannelDataBlock-related handlers
#===============================================================================

class SimpleChannelDataBlock(BaseDataBlock):
    """ Wrapper for SimpleChannelDataBlock elements, which consist of only a
        binary payload of raw data prefixed with a 16b timestamp and an 8b
        channel ID. Also keeps track of some metadata used by its Channel.
        
        :cvar headerSize: The size of the header information stored in the
            element's payload, if any.
    """
    headerParser = struct.Struct(">HB")
    headerSize = headerParser.size
    maxTimestamp = 2**16

    def __init__(self, element):
        super(SimpleChannelDataBlock, self).__init__(element)
        self.element = element
        self.startTime, self.channel = self.getHeader()
        self.payloadSize = element.size - self.headerSize
   
    
    def __len__(self):
        """ x.__len__() <==> len(x)
            This returns the length of the payload.
        """
        return self.numSamples
    
    
    def getHeader(self):
        """ Extract the block's header info. In SimpleChannelDataBlocks,
            this is part of the payload.
        """
        return self.headerParser.unpack_from(self.payload)
    

    @property
    def timestamp(self):
        return self.getHeader()[0]


    def parseWith(self, parser, start=None, end=None, step=1, subchannel=None):
        """ Parse an element's payload. Use this instead of directly using
            `parser.parse()` for consistency's sake.
            
            :param parser: The DataParser to use
            :keyword start: First subsample index to parse 
            :keyword end: Last subsample index to parse
            :keyword step: The number of samples to skip, if the start and end
                cover more than one sample.
            :keyword subchannel: The subchannel to get, if specified.
        """
        # SimpleChannelDataBlock payloads contain header info; skip it.
        data = self.payload
        start, end, step = slice(start, end, step).indices(
            (self.payloadSize-self.headerSize) // parser.size
        )

        start = self.headerSize + (start*parser.size)
        end = self.headerSize + (end*parser.size)
        
        parser_unpack_from = parser.unpack_from
        if subchannel is not None:
            for i in range(start, end, parser.size*step):
                yield parser_unpack_from(data, i)[subchannel]
        else:
            for i in range(start, end, parser.size*step):
                yield parser_unpack_from(data, i)


    def parseByIndexWith(self, parser, indices, subchannel=None):
        """ Parse an element's payload and get a specific set of samples. Used
            primarily for resampling tricks.
            
            :param parser: The DataParser to use
            :param indices: A list of indices into the block's data. 
            :keyword subchannel: The subchannel to get, if specified.
        """
        # SimpleChannelDataBlock payloads contain header info; skip it.
        data = self.payload
        parser_unpack_from = parser.unpack_from
        parser_size = parser.size
        for i in indices:
            if i >= self.numSamples:
                continue
            idx = self.headerSize + (i*parser_size)
            if subchannel is not None:
                yield parser_unpack_from(data, idx)[subchannel]
            else:
                yield parser_unpack_from(data, idx)



class SimpleChannelDataBlockParser(ElementHandler):
    """ 'Factory' for SimpleChannelDataBlock elements. Instantiated once
        per session (or maybe channel, depending). It handles the modulus
        correction of the block's short timestamps.
        
        :cvar elementName: The name of the element handled by this parser
        :cvar product: The class of object generated by the parser
    """
    product = SimpleChannelDataBlock
    elementName = product.__name__
    isHeader = False

    # The default block time scalar.
    timeScalar = 1000000.0 / 2**15

    def __init__(self, doc, **kwargs):
        super(SimpleChannelDataBlockParser, self).__init__(doc, **kwargs)

        # Timestamp conversion/correction is done per channel        
        self.timestampOffset = {}
        self.lastStamp = {}
        self.timeScalars = {}
        self.timeModulus = {}


    def fixOverflow(self, block, timestamp):
        """ Return an adjusted, scaled time from a low-resolution timestamp.
        """
        # TODO: Identify blocks with non-modulo timestamps and just return the
        #    unmodified timestamp. Will be slightly more efficient.
        
        channel = block.getHeader()[1]
        modulus = self.timeModulus.setdefault(channel, block.maxTimestamp)
        offset = self.timestampOffset.setdefault(channel, 0)
        
        if timestamp > modulus:
            # Timestamp is (probably) not modulo; will occur in split files.
            offset = timestamp - (timestamp % modulus)
            timestamp = timestamp % modulus
            self.timestampOffset[channel] = offset
        elif timestamp < self.lastStamp.get(channel, 0):
            # Modulo rollover (probably) occurred.
            offset += modulus
            self.timestampOffset[channel] = offset
            
        self.lastStamp[channel] = timestamp
        timestamp += self.timestampOffset[channel]
        return timestamp * self.timeScalars.setdefault(channel, self.timeScalar)
    
   
    def parse(self, element, sessionId=None, timeOffset=0):
        """ Create a (Simple)ChannelDataBlock from the given EBML element.
        
            :param element: A sample-carrying EBML element.
            :keyword sessionId: The session currently being read; defaults to
                whatever the Dataset says is current.
            :return: The number of subsamples read from the element's payload.
        """
        try:
            block = self.product(element)
            timestamp, channel = block.getHeader()
        except struct.error as e:
            raise ParsingError("Element would not parse: %s (ID %d) @%d (%s)" %
                               (element.name, element.id, element.offset, e))
        except AttributeError:
            # Can happen if the block had no timestamp (broken imported data?)
            # TODO: Actually handle, instead of ignoring?
            logger.warning("XXX: bad attribute in element %s" % element)
            return 0
            
        
        block.startTime = timeOffset + int(self.fixOverflow(block, timestamp))
        if block.endTime is not None:
            block.endTime = timeOffset + int(self.fixOverflow(block, block.endTime))

        if channel not in self.doc.channels:
            # Unknown channel; could be debugging info, so that might be okay.
            # FUTURE: Better handling of unknown channel types. Low priority.
            return 0

        try:
            ch = self.doc.channels[channel]
            ch.getSession(sessionId).append(block)
            return block.getNumSamples(ch.parser) * len(ch.children)
        except ZeroDivisionError:
            return 0


#===============================================================================
# ChannelDataBlock: Element wrapper and handler
#===============================================================================

class ChannelDataBlock(BaseDataBlock):
    """ Wrapper for ChannelDataBlock elements, which features additional data
        excluded from the simple version. ChannelDataBlock elements are 'master'
        elements with several child elements, such as full timestamps and
        and sample minimum/mean/maximum.
    """
    maxTimestamp = 2**24

    def __init__(self, element):
        super(ChannelDataBlock, self).__init__(element)
        self._payloadIdx = None
        self._payloadEl = None
        
        self._minMeanMaxEl = None
        self._minMeanMax = None
        
        self.element = element
        for el in element:
            # These are roughly in order of probability, optional and/or
            # unimplemented elements are at the end.
            if el.name == "ChannelIDRef":
                self.channel = el.value
            elif el.name == "ChannelDataPayload":
                self._payloadEl = el
                self.payloadSize = el.size
            elif el.name == "StartTimeCodeAbsMod":
                self.startTime = el.value
                self._timestamp = el.value
            elif el.name == "EndTimeCodeAbsMod":
                self.endTime = el.value
            elif el.name == "ChannelDataMinMeanMax":
#                 self.minMeanMax = el.value
                self._minMeanMaxEl = el
            elif el.name == "Void":
                continue
            elif el.name == 'Attribute':
                parseAttribute(self, el)
                el.gc()
            elif el.name == "StartTimeCodeAbs":
                # TODO: store indicator that the start timestamp is non-modulo?
                self.startTime = el.value
                self._timestamp = el.value
            elif el.name == "EndTimeCodeAbs":
                # TODO: store indicator that the end timestamp is non-modulo?
                self.endTime = el.value
            elif el.name == "ChannelFlags":
                # FUTURE: Handle channel flag bits
                continue
            # Add other child element handlers here.
        
        element.gc(recurse=False)
        
        # Single-sample blocks have a total time of 0. Old files did not write
        # the end timestamp; if it's missing, duplicate the starting time.
        if self.endTime is None:
            self.endTime = self.startTime
    
    
    @property
    def payload(self):
        return self._payloadEl.value
    
    
    @property
    def minMeanMax(self):
        if self._minMeanMaxEl is None:
            return self._minMeanMax
        return self._minMeanMaxEl.value
    
    
    @minMeanMax.setter
    def minMeanMax(self, v):
        # Explicitly set (done when block contains a single sample)
        self._minMeanMax = v
    
    
    def getHeader(self):
        """ Extract the block's header info (timestamp and channel ID).
        """
        return self._timestamp, self.channel


class ChannelDataArrayBlock(ChannelDataBlock):
    def __init__(self, element):
        super(ChannelDataArrayBlock, self).__init__(element)
        self._payload = None

        self._parser = None
        self._streamDtype = None
        self._commonDtype = None

    @property
    def payload(self):
        if self._payload is None:
            self._payload = np.array(self._payloadEl.value)
            self._payloadEl.gc()
        return self._payload

    # Define standard mapping from struct to numpy typestring
    #   (conversions taken from struct & numpy docs:)
    #   https://docs.python.org/3/library/struct.html#format-characters
    #   https://numpy.org/doc/stable/reference/arrays.dtypes.html#specifying-and-constructing-data-types
    TO_NP_TYPESTR = {
        # 'x': '',
        'c': 'b',
        'b': 'b',
        'B': 'B',
        '?': '?',
        'h': 'i2',
        'H': 'u2',
        'i': 'i4',
        'I': 'u4',
        'l': 'i4',
        'L': 'u4',
        'q': 'i8',
        'Q': 'u8',
        # 'n': '',
        # 'N': '',
        # 'e': 'f2',  unsupported in Python3.5
        'f': 'f4',
        'd': 'f8',
        # 's': '',
        # 'p': '',
        # 'P': '',
    }

    def parseWith(self, parser, start=None, end=None, step=1, subchannel=None):
        """ Parse an element's payload. Use this instead of directly using
            `parser.parse()` for consistency's sake.

            :param parser: The DataParser to use
            :keyword start: First subsample index to parse
            :keyword end: Last subsample index to parse
            :keyword step: The number of samples to skip, if the start and end
                cover more than one sample.
            :keyword subchannel: The subchannel to get, if specified.
        """

        # All numpy-compatible data streams use `struct.Struct` parsers
        isNumpyCompatibleFormat = isinstance(parser, struct.Struct)

        if not isNumpyCompatibleFormat:
            # No parser format -> assume that data is from an old .ide file
            #   & should be handled safely using the parser object
            blocks = list(ChannelDataBlock.parseWith(
                self, parser, start, end, step, subchannel
            ))
            return np.array(blocks, dtype=np.float64).T

        if parser is self._parser:
            streamDtype = self._streamDtype
            commonDtype = self._commonDtype
        else:
            parser_format = parser.format
            if isinstance(parser_format, (bytes, bytearray)):
                parser_format = parser_format.decode()
            if parser_format[0] in ['<', '>']:
                endian = parser_format[0]
                parser_format = parser_format[1:]
            else:
                endian = '>'

            streamDtype = np.dtype(','.join([
                endian + self.TO_NP_TYPESTR[typeId] for typeId in parser_format
            ]))

            isHomogeneous = len(set(parser_format)) == 1
            if isHomogeneous:
                commonDtype = np.dtype(endian + self.TO_NP_TYPESTR[parser_format[0]])
            else:
                commonDtype = None

            self._parser = parser
            self._streamDtype = streamDtype
            self._commonDtype = commonDtype

        if commonDtype is not None:
            rawData = np.frombuffer(self.payload, dtype=commonDtype)
            rawData = rawData.reshape(-1, len(streamDtype) or 1)[start:end:step].T

            if len(streamDtype) > 0 and subchannel is not None:
                return rawData[[subchannel]]
            return rawData

        rawData = np.frombuffer(self.payload, dtype=streamDtype)[start:end:step]

        # Special cases for single-channel outputs
        if subchannel is not None:
            return rawData[streamDtype.names[subchannel]][np.newaxis]

        data = np.empty((len(streamDtype),) + rawData.shape,
                        dtype=np.float64)
        for i, chName in enumerate(streamDtype.names):
            data[i] = rawData[chName]

        return data

    def parseByIndexWith(self, parser, indices, subchannel=None):
        """ Parse an element's payload and get a specific set of samples. Used
            primarily for resampling tricks.

            :param parser: The DataParser to use
            :param indices: A list of indices into the block's data.
            :keyword subchannel: The subchannel to get, if specified.
        """
        # SimpleChannelDataBlock payloads contain header info; skip it.
        return self.parseWith(parser, subchannel=subchannel)[:, indices]

    def parseMinMeanMax(self, parser):
        if self.minMeanMax is None:
            return None
        try:
            stats = np.array([
                parser.unpack_from(self.minMeanMax, parser.size*i)
                for i in range(3)
            ])
            # NOTE: Because the SSX Z axis is inverted, the min/max need correction
            # This may create a performance hit. Optimize?
            idx = (stats[0] > stats[2])
            stats[[0, 2]][..., idx] = stats[[2, 0]][..., idx]

            self.min, self.mean, self.max = [i.copy() for i in stats]
            return stats
        except struct.error:
            # Bad min/mean/max data: too short. Ignore it.
            self.minMeanMax = None
            return None


class ChannelDataBlockParser(SimpleChannelDataBlockParser):
    """ 'Factory' for ChannelDataBlock elements. Instantiated once per 
        session (or maybe channel, depending). It handles the modulus
        correction of the block's timestamps.
        
        :cvar elementName: The name of the element parsed by this parser
        :cvar product: The class of object generated by the parser
    """
    product = ChannelDataBlock
    elementName = product.__name__

    timeScalar = 1000000.0 / 2**15


class ChannelDataArrayBlockParser(ChannelDataBlockParser):
    """ Factory for ChannelDataArrayBlock elements.  Instantiated once per
        session/channel, handles modulus correction for the blocks' timestamps.
        Unlike the ChannelDataBlockParser, this returns blocks which store
        (cache?) data as numpy arrays.

        :cvar product: The class of object generated by the parser
    """
    product = ChannelDataArrayBlock
    elementName = product.__name__ # product.__name__


################################################################################
#===============================================================================
#--- RecordingProperties element and sub-element parsers
#===============================================================================
################################################################################

class RecorderPropertyParser(ElementHandler):
    """ Base class for elements that just add a value to the Dataset's
        `recorderInfo` but aren't in the `RecorderInfo` element.
    """
    isHeader = True
    isSubElement = True
    
    def parse(self, element, **kwargs):
        if self.doc is not None:
            if element.name == 'Attribute':
                parseAttribute(self.doc, element)
            else:
                self.doc.recorderInfo[element.name] = element.value
   

#===============================================================================
# RecordingProperties: Calibration
#===============================================================================

class PolynomialParser(ElementHandler):
    """ The handler for both Univariate and Bivariate calibration polynomials.
        Each are a subclass of this, although this class does all the work.
    """
    elementName = ("UnivariatePolynomial", "BivariatePolynomial")
    isSubElement = True
    isHeader = True
    
    # Parameter names: mapping of element names to the keyword arguments used
    # to instantiate a polynomial object. Also used to remove unknown elements
    # (see `renameKeys`).
    parameterNames = {"CalID": "calId",
                      "CalReferenceValue": "reference",
                      "BivariateCalReferenceValue": "reference2",
                      "BivariateChannelIDRef": "channelId",
                      "BivariateSubChannelIDRef": "subchannelId",
                      "PolynomialCoef": "coeffs",
                      "Attribute": "attributes"}

    def parse(self, element, **kwargs):
        """
        """
        # Element name (plus ID and file position) for error messages
        elName = self.getElementName(element)
        params = renameKeys(element.dump(), self.parameterNames)
        params['dataset'] = self.doc
        
        coeffs = params.pop("coeffs", None)
        if coeffs is None:
            raise ParsingError("%s had no coefficients" % elName)

        if "calId" not in params:
            raise ParsingError("%s had no calibration ID" % elName)
        
        if element.name == "BivariatePolynomial":
            # Bivariate polynomial. Do extra error checking.
            if "channelId" not in params or "subchannelId" not in params:
                raise ParsingError("%s had no channel reference!" % elName)
            if len(coeffs) != 4:
                raise ParsingError("%s supplied %d coefficients; 4 required" %
                                   (elName, len(coeffs)))
            cal = transforms.Bivariate(coeffs, **params)
            
        elif element.name == "UnivariatePolynomial":
            cal = transforms.Univariate(coeffs, **params)
            
        else:
            # Unknown polynomial type. 
            raise ParsingError("%s: unknown polynomial type" % elName)
        
        # self.doc might (validly) be None if a configuration tool is 
        # reading the device info file, rather than reading a recording file. 
        if self.doc is not None:
            self.doc.addTransform(cal)
        
        return cal


class CalibrationElementParser(RecorderPropertyParser):
    """ Simple handler for calibration birthday (optional). """
    elementName = ("CalibrationDate",
                   "CalibrationExpiry",
                   "CalibrationSerialNumber")


class CalibrationListParser(ElementHandler):
    """ Root-level parser for calibration data. Handles parsing of the
        individual calibration elements (its children). Unlike (most) other
        parsers, this one can be instantiated without a reference to a 
        `dataset.Dataset`. It also keeps a copy of all the calibration items
        in a `items` attribute (a list). 
    """
    isHeader = True
    elementName = "CalibrationList"
    children = (PolynomialParser, CalibrationElementParser)


#===============================================================================
# RecordingProperties: Sensor-related parsers
#===============================================================================

class SensorListParser(ElementHandler):
    """ Handle `SensorList` elements, creating the individual Sensors from
        the element's children.
    """
    elementName = "SensorList"
    isSubElement = True

    # Parameter names: mapping of element names to the keyword arguments used
    # to instantiate the various children of SensorListParser. Also used to 
    # remove unknown elements (see `renameKeys`).
    parameterNames = {
        "Sensor": "sensors",
        "SensorID": "sensorId",
        "SensorName": "name",
        "TraceabilityData": "traceData",
        "SensorSerialNumber": "serialNum",
        "Attribute": "attributes",
#         "SensorBwLimitIDRef": "bandwidthLimitId" # FUTURE
    }
    
    def parse(self, element, **kwargs):
        """ Parse a SensorList 
        """
#         data = parse_ebml(element.value)
        data = element.dump()
        if 'attributes' in data:
            atts = self.doc.recorderInfo.setdefault('sensorAttributes', {})
            atts.update(decode_attributes(data['attributes']))
        data = renameKeys(data, self.parameterNames)
        if "sensors" in data:
            for sensor in data['sensors']:
                self.doc.addSensor(**sensor)
    

#===============================================================================
# RecordingProperties: Channel and Subchannel parsers
#===============================================================================

class ChannelParser(ElementHandler):
    """ Handle individual Channel elements. Separate from ChannelList so it can
        be subclassed for Plots. 
    """
    elementName = "Channel"
    isSubElement = True
    isHeader = True

    parameterNames = {
        # Parent `Channel parameters.
        "ChannelID": "channelId",
        "ChannelName": "name",
        # "TimeCodeScale": "timeScale",
        # "TimeCodeModulus": "timeModulus",
        "SampleRate": "sampleRate",
        "ChannelParser": "parser",
        "ChannelFormat": "format",
        "ChannelCalibrationIDRef": "transform",
        "TimeCodeScale": "timeScalar",
        "TimeCodeModulus": "timeMod",
        "cache": "cache",
        "singleSample": "singleSample",
        "SubChannel": "subchannels", # Multiple, so it will parse into a list

        # Child SubChannel parameters; `renameKeys` operates recursively.
        "SubChannelID": "subchannelId",
        "SubChannelName": "name",
        "SubChannelAxisName": "axisName",
        "SubChannelCalibrationIDRef": "transform",
        "SubChannelLabel": "label",
        "SubChannelUnits": "units",
        "SubChannelRangeMin": "rangeMin",
        "SubChannelRangeMax": "rangeMax",
        "SubChannelSensorRef": "sensorId",
        "SubChannelWarningRef": "warningId",
        "SubChannelPlotColor": "color",
        
        # Generic attribute elements.
        "Attribute": "attributes"
    }
    
    
    def parse(self, element, **kwargs):
        """ Create the `dataset.Channel` object and its `dataset.SubChannel` 
            children elements from a `Channel` element of the `ChannelList`.
        """
        data = renameKeys(element.dump(), self.parameterNames)
        
        # Pop off the subchannels; create them in a second pass.
        subchannels = data.pop('subchannels', None)
        
        channelId = data['channelId']
        
        # Parsing. Either a regular struct.Struct, or a custom parser.
        if 'parser' in data:
            # A named parser; use the special-case parser function.
            data.pop('format', None)
            data['parser'] = DATA_PARSERS[data['parser']]()
        elif 'format' in data:
            # build struct instead.
            # TODO (future): Handle special (non-standard) format characters
            data['parser'] = struct.Struct(data.pop('format'))

        # sampleRate is stored as a string to avoid floats on the recorder.
        if 'sampleRate' in data:
            data['sampleRate'] = valEval(data['sampleRate'])
        
        # Channel timestamp correction stuff.
        timeScale = data.pop('timeScalar', None)
        timeModulus = data.pop('timeMod', None)
        
        # Update timestamp modulo in parsers
        for p in [x for x in list(self.doc._parsers.values()) if x.makesData()]:
            if timeModulus is not None:
                p.timeModulus[channelId] = timeModulus
            if timeScale is not None:
                p.timeScalars[channelId] = valEval(timeScale) * 1000000.0
        
        ch = self.doc.addChannel(**data)
        
        if subchannels is not None:
            for subData in subchannels:
                displayRange = subData.pop("rangeMin", None), subData.pop("rangeMax", None)
                subData['displayRange'] = None if None in displayRange else displayRange
                units = subData.pop('label', None), subData.pop('units', None)
                if units[0] is None:
                    units = units[1], units[1]
                subData['units'] = None if None in units else units
                
                ch.addSubChannel(**subData)
        return ch


class PlotListParser(ChannelParser):
    """ Handle the parent of all `Plot` elements. 
        Note: Not currently implemented.
    """
    elementName = "PlotList"
    isSubElement = True

    # Most Plot parameters are the same as those of Subchannel. 
    parameterNames = {
        'PlotSource': 'sources',
        'PlotChannelRef': 'channelId',
        'PlotSubChannelRef': 'subchannelId'
    }
    
    def __init__(self, *args, **kwargs):
        pnames = super(PlotListParser, self).parameterNames.copy()
        self.parameterNames.update(pnames)
        self.parameterNames = pnames

    def parse(self, element, **kwargs):
        """
        """
        # FUTURE: IMPLEMENT, but later. Not required for immediate goals.
        pass
    

class ChannelListParser(ElementHandler):
    """ Handle `ChannelList` elements and their children. Just wraps the
        sub-element parsers.
    """
    elementName = "ChannelList"
    isSubElement = True
    isHeader = True
    children = (ChannelParser,) #PlotListParser)


#===============================================================================
# 
#===============================================================================

class WarningListParser(ElementHandler):
    """ Handle `WarningList` elements and their children (`Warning`).
    """
    elementName = "WarningList"
    isSubElement = True
    isHeader = True
    
    # Note: Includes child `Warning` parameters.
    parameterNames = {
        "Warning": "warnings",
        "WarningID": "warningId",
        "WarningChannelRef": "channelId",
        "WarningSubChannelRef": "subchannelId",
        "WarningRangeMin": "low",
        "WarningRangeMax": "high",
        "Attribute": "attributes"
    }
    
    def parse(self, element, **kwargs):
        raw = element.dump()
        data = renameKeys(raw, self.parameterNames)
        
        if 'warnings' in data:
            for w in data['warnings']:
                self.doc.addWarning(**w)
    
    
#===============================================================================
# RecordingProperties: RecorderInfo
#===============================================================================

class RecorderInfoParser(ElementHandler):
    """ Handle the `RecorderInfo` element, child of RecordingProperties.
    """
    elementName = "RecorderInfo"
    isSubElement = True
    
    def parse(self, element, **kwargs):
        """
        """
        # This one is simple; it just sticks the data into the Dataset.
        val = element.dump()
        if 'Attribute' in val:
            self.doc.recorderInfo.update(decode_attributes(val.pop('Attribute')))
        self.doc.recorderInfo.update(val)
        return self.doc.recorderInfo  


#===============================================================================
# 
#===============================================================================

class BwLimitListParser(ElementHandler):
    """ Handle the `BwLimitList` element, child of RecordingProperties.
    """
    elementName = "BwLimitList"
    isSubElement = True
    isHeader = True

    def parse(self, element, **kwargs):
        """
        """
        self.doc.bandwidthLimits = limits = {}
        val = element.value.dump()
        for limit in val.get('BwLimit', []):
            limitId = limit.pop('BwLimitID', None)
            if limitId is not None:
                limits[limitId] = limit
        return limits  

    
#===============================================================================
# RecordingProperties parent element parser.
#===============================================================================

class RecordingPropertiesParser(ElementHandler):
    """ Stub for RecordingProperties element handler. All relevant data is in
        its child elements.
    
        :cvar elementName: The name of the element handled by this parser
    """
    elementName = "RecordingProperties"
    isSubElement = False
    isHeader = True
    children = (RecorderInfoParser, 
                SensorListParser, 
                ChannelListParser, 
                WarningListParser)


################################################################################
#===============================================================================
#--- Miscellaneous element parsers
#===============================================================================
################################################################################

class TimeBaseUTCParser(ElementHandler):
    """ Handle TimeBaseUTC elements, applying it as the UTC start time of the
        current Session.
    """
    elementName = "TimeBaseUTC"
    isHeader = True
    
    def parse(self, element, **kwargs):
        val = element.value
        self.doc.lastUtcTime = val
        self.doc.lastSession.utcStartTime = val


class RecorderUserDataParser(ElementHandler):
    """
    """
    elementName = "RecorderUserData"
    isSubElement = True
    isHeader = True
    
    parameterNames = {"RecorderName": "RecorderName",
                      "RecorderDesc": "RecorderDescription"}

    def parse(self, element, **kwargs):
        if self.doc is not None:
            raw = element.dump()
            data = renameKeys(raw, self.parameterNames)
            self.doc.recorderInfo.update(data)


class RecorderConfigurationListParser(ElementHandler):
    """
    """
    elementName = "RecorderConfigurationList"
    isSubElement = True
    isHeader = True

    parameterNames = {0x8ff7f: "RecorderName",
                      0x9ff7f: "RecorderDescription"}
    
    def parse(self, element, **kwargs):
        if self.doc is not None:
            for el in element:
                if el[0].name != "ConfigID":
                    continue
                if el[0].value in self.parameterNames:
                    name = self.parameterNames[el[0].value]
                    self.doc.recorderInfo[name] = el[1].value


class RecorderConfigurationParser(ElementHandler):
    """ Handle Recorder configuration data in a recording. This just parses it
        and stores it verbatim.
    """
    elementName = "RecorderConfiguration"
    isHeader = True
    isSubElement = False
    children = (RecorderConfigurationListParser, RecorderUserDataParser)


class AttributeParser(ElementHandler):
    """ Handle a root-level Attribute element. 
    """
    elementName = "Attribute"
    isHeader = True
    isSubElement = False
     
    def parse(self, element, **kwargs):
        parseAttribute(self.doc, element)
        return 0

    
#===============================================================================
# 
#===============================================================================

class NullParser(ElementHandler):
    """ Dummy handler for all root level elements we know of but don't care 
        about (at least for now). Note that this includes the `EBML` element,
        because the EBML library itself has already handled it.
    """
    elementName = ('EBML',
                   'ElementTag',
                   'RecorderConfigurationList',
                   'Session',
                   'Sync',
                   'Void',
                   'UnknownElement')

    def parse(self, *args, **kwargs):
        pass



"""
Calibration and Transform classes. These are callable objects that modify
data. Functionally, calibration is the same as a Transform, the difference
being in how they are used.

:author: dstokes
"""

__all__ = ['Transform', 'Univariate', 'Bivariate', 'CombinedPoly', 'PolyPoly',
           'AccelTransform']

from collections import OrderedDict
import math
from time import sleep

import logging
logger = logging.getLogger('idelib-archive')
logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s")

#===============================================================================
# 
#===============================================================================


class Transform(object):
    """ Base class for all data-manipulating objects (e.g. calibration
        polynomials). Instantiates as a function-like object representing any 
        processing that an event requires, including basic calibration at the 
        low level and adjustments for display at the high level.
    """
    modifiesTime = False
    modifiesValue = False
    units = None
    
    def __init__(self, *args, **kwargs):
        self.id = None
        self._str = "x"
        self._source = "lambda x: x"
        self._function = eval(self._source, {'math': math})
        self._lastSession = None
        self._timeOffset = 0
        
        # Custom attributes, e.g. Attribute elements in the EBML.
        self.attributes = kwargs.pop('attributes', None)
    
    
    def copy(self):
        """ Create a duplicate of this Transform.
        """
        t = self.__class__()
        for attr in ('id', '_str', '_source', '_function', '_lastSession', 
                     '_timeOffset'):
            setattr(t, attr, getattr(self, attr, None))
        return t
    
    
    def __str__(self):
        return self._str
    
    
    def __repr__(self):
        cname = self.__class__.__name__
        if self.id is None:
            return "<%s: (%s)>" % (cname, self._str)
        return "<%s (ID %d): (%s)>" % (cname, self.id, self._str)


    def __hash__(self):
        return hash(self._str)


    def __eq__(self, other):
        try:
            return self._str == other._str
        except AttributeError:
            return False
    
    
    def __ne__(self, other):
        return not self.__eq__(other)


    @property
    def function(self):
        """ The generated polynomial function itself. """
        return self._function


    @property
    def source(self):
        """ The optimized source code of the polynomial. """
        return self._source
    
    
    def __call__(self, timestamp, value, session=None, noBivariates=False):
        if session != self._lastSession:
            self._timeOffset = 0 if session.startTime is None else session.startTime
            self._session = session
        return timestamp + self._timeOffset, self._function(value)


    def isValid(self, session=None, noBivariates=False):
        """ Check the validity of the Transform.
        """
        # TODO: More base validity tests for all Transform subclasses?
        try:
            return (self.function is not None)
        except AttributeError:
            return False

    
    #===========================================================================
    # 
    #===========================================================================
    
    @classmethod
    def null(cls, *args, **kwargs):
        return args[0]


#===============================================================================
# Simple Transforms. These could be represented by Univariate polynomials,
# but are hard-coded for efficiency's sake.
#===============================================================================


class AccelTransform(Transform):
    """ A simple transform to convert accelerometer values (parsed as
        int16) to floats in the range -100 to 100 G.
         
        This assumes that the data was parsed by `AccelerometerParser`, which
        puts the raw values in the range -32768 to 32767.
    """
    modifiesValue = True
     
    def __init__(self, amin=-100, amax=100, calId=0, dataset=None):
        self.id = calId
        self.range = (amin, amax)
        self._str = "(x / 32767.0) * %.3f" % amax
        self._source = "lambda x: x * %f" % (amax / 32767.0)
        self._function = eval(self._source, {'math': math})
        self._lastSession = None
        self._timeOffset = 0
        
        self.references = (0,)
        self.coefficients = ((amax / 32767.0), amax)

        self.dataset = None


    def copy(self):
        """ Create a duplicate of this Transform.
        """
        t = self.__class__(self.range[0], self.range[1], self.id, self.dataset)
        return t


#===============================================================================
# Polynomial Generators
#===============================================================================


class Univariate(Transform):
    """ A simple calibration polynomial in the general form:
        `y=(coeffs[0]*(x**n))+(coeffs[1]*(x**n-1))+...+(coeffs[n]+(x**0))`.
        
        Instances are function-like objects that take one argument: a sensor
        reading.
    """
    modifiesValue = True
    
    @classmethod
    def _floatOrInt(cls, v):
        """ Helper method to convert floats with no decimal component to ints. """
#         iv = int(v)
#         return iv if iv == v else v
        # Not a good optimization. Only convert 0 to integer.
        return 0 if v == 0 else float(v)
    
    
    @classmethod
    def _stremove(cls, s, old):
        """ Helper method to remove a set of substrings from a string. """
        result = str(s)
        for o in old:
            result = result.replace(o, '')
        return result
    
    
    @classmethod
    def _streplace(cls, s, *args):
        """ Helper method to replace multiple substrings. """
        for old, new in args:
            s = s.replace(old, new)
        return s
    
    
    @classmethod
    def _fixSums(cls, s):
        """ Helper method to replace consecutive addition/subtraction combos. """
        result = str(s)
        for old, new in (("--", "+"), ("-+", "-"), ("+-", "-"), ("++", "+")):
            result = result.replace(old, new)
        return result.lstrip('+')
    
    
    # XXX: kwargs added to __init__() as work-around for bad SSS templates! 
    def __init__(self, coeffs, calId=None, dataset=None, reference=0, 
                 varName="x", attributes=None, **kwargs):
        """ Construct a simple polynomial function from a set of coefficients.
            
            :param coeffs: A list of coefficients
            :keyword calId: The polynomial's calibration ID (if any).
            :keyword dataset: The parent `dataset.Dataset`.
            :keyword reference: A reference value to be subtracted from the
                variable.
            :keyword varName: The name of the variable to be used in the
                string version of the polynomial. For display purposes.
            :keyword attributes: A dictionary of generic attributes, e.g.
                ones parsed from `Attribute` EBML elements, or `None`.
        """
        self.id = calId
        self.dataset = dataset
        self._coeffs = tuple(coeffs)
        self._variables = (varName,)
        self._references = (reference,)
        self._session = None
        self._lastSession = None
        self._timeOffset = 0
        
        self.attributes = attributes
        
        self._build()
        

    def __eq__(self, other):
        try:
            return self._coeffs == other._coeffs and self._references == other._references
        except AttributeError:
            return False


    def copy(self):
        """ Create a duplicate of this Transform.
        """
        return self.__class__(self._coeffs, calId=self.id, dataset=self.dataset,
                              reference=self._references[0], 
                              varName=self._variables[0])

    
    def _build(self):
        """ Internal method that (re-)constructs the polynomial function.
        """
        varName = str(self._variables[0])
        srcVarName = "x"
        reference = self._references[0]
        coeffs = self._coeffs
        
        if reference != 0:
            varName = "(%s-%s)" % (varName, reference)
            srcVarName = "(%s-%s)" % (srcVarName, reference)
        
        # f is used to build the lambda
        # strF is used to build the string version
        coeffs = list(reversed(coeffs))
        f = [coeffs[0]] if coeffs[0] != 0 else []
        strF = f[:]
        coeffs = [self._floatOrInt(c) for c in coeffs]
        for p, v in enumerate(coeffs[1:], 1):
            # optimization: x*0 == 0
            if v == 0:
                continue
    
            # optimization: pow() is more expensive than lots of multiplication
            x = "*".join([srcVarName]*p)
            strX = "pow(%s,%s)" % (varName, p) if p > 1 else varName
    
            # optimization: remove multiplication by 1
            q = "%s*" % v if v != 1 else ""
    
            f.append("(%s%s)" % (q, x))
            strF.append("(%s%s)" % (q, strX))
    
        self._str = "+".join(map(str, reversed(strF)))
        self._str = self._fixSums(self._str)
        
        self._source = "lambda x: %s" % ("+".join(map(str, reversed(f))))
        self._source = self._fixSums(self._source)
        self._function = eval(self._source, {'math': math})
    
    
    @property
    def coefficients(self):
        """ The polynomial's coefficients. """
        return self._coeffs
    
    
    @coefficients.setter
    def coefficients(self, val):
        self._coeffs = val
        self._build()
        
        
    @property
    def variables(self):
        """ The name(s) of the variable(s) used in the polynomial. """
        return self._variables
    
    
    @variables.setter
    def variables(self, val):
        self._variables = val
        self._build()
    
    
    @property
    def references(self):
        """ The constant offset(s). """
        return self._references
    
    
    @references.setter
    def references(self, val):
        self._references = val
        self._build()
    
    
    def asDict(self):
        """ Dump the polynomial as a dictionary. Intended for use when
            generating EBML.
        """
        return OrderedDict((('CalID', self.id),
                            ('CalReferenceValue', self._references[0]),
                            ('PolynomialCoef', self._coeffs)))


#===============================================================================
# 
#===============================================================================

class Bivariate(Univariate):
    """ A two-variable polynomial in the general form
        `v=(A*x*y)+(B*x)+(C*y)+D`.
        
        Instances are function-like objects that take one argument: a sensor
        reading time and value.
    """
    
    def __init__(self, coeffs, calId=None, dataset=None, reference=0, 
                 reference2=0, channelId=None, subchannelId=None, varNames="xy",
                 attributes=None):
        """ Construct the two-variable polynomial.
            
            :param coeffs: A list of coefficients. Must contain 4!
            :keyword reference: A reference value to be subtracted from the
                'x' variable.
            :keyword reference2: A reference value to be subtracted from the
                'y' variable.
            :keyword varNames: The names of the variables to be used in the
                string version of the polynomial. For display purposes; they
                can be any arbitrary (but hopefully meaningful) strings.
            :keyword calId: The polynomial's calibration ID (if any).
            :keyword dataset: The parent `dataset.Dataset`.
            :keyword attributes: A dictionary of generic attributes, e.g.
                ones parsed from `Attribute` EBML elements, or `None`.
        """
        self.dataset = dataset
        self._eventlist = None
        self._sessionId = None
        self.channelId, self.subchannelId = channelId, subchannelId
        if channelId is None or subchannelId is None:
            raise ValueError("Bivariate polynomial requires channel and "
                    "subchannel IDs; got %r, %d" % (channelId, subchannelId))
        
        if len(coeffs) != 4:
            raise ValueError("Bivariate polynomial must have exactly 4 "
                             "coefficients; %d were supplied" % len(coeffs))
        if len(varNames) != 2:
            raise ValueError("Bivariate polynomial must have two variable "
                             "names; %d were supplied" % len(varNames))
        
        self.id = calId
        
        self._references = (float(reference), float(reference2))
        self._coeffs = tuple(map(float, coeffs))
        self._variables = tuple(map(str, varNames))
        
        self._session = None
        self._timeOffset = 0
        
        self.attributes = attributes
        
        self._build()
        
        
    def __eq__(self, other):
        try:
            return self._str == other._str
        except AttributeError:
            return False
    
    
    def copy(self):
        """ Create a duplicate of this Transform.
        """
        # This could be optimized by circumventing the polynomial rebuild,
        # and instead using the already-generated source.
        return self.__class__(self._coeffs, dataset=self.dataset, 
               channelId=self.channelId, subchannelId=self.subchannelId, 
               reference=self._references[0], reference2=self._references[1], 
               varNames=self._variables, calId=self.id)

    
    @classmethod
    def _reduce(cls, src):
        """ Simple reduction for polynomial expressions, removing redundant
            parts (adding/subtracting 0, multiplying by 0 or 1, etc.).
        """
        old = None
        while old != src:
            old = src
            src = cls._streplace(src, 
                ('(0+', '('), ('(0.0+', '('), ('(0-', '(-'), ('(0.0-', '(-'),
                ('(0*x', '(0'), ('(0.0*x', '(0'), 
                ('(0*y', '(0'), ('(0.0*y', '(0'),
                ('(0*x*y)+', ''), ('(0*x)+', ''), ('(0*y)+', ''),
                ('*)', ')'), ('+)', ')'), 
                ('(0)', ''), ('(0.0)', ''), ('()', ''), 
                ('(1)', '1.0'), ('(1.0)', '1.0'),
                ("(1*", "("), ("(1.0*", "("),  # ("(x)", "x"), ("(y)", "y"),
                ("--", "+"), ("-+", "-"), ("+-", "-"), ("++", "+")
            )
            if src.endswith('+0') or src.endswith('+0.0'):
                src = src.rstrip('+0.')
        return src
    
    
    def _build(self):
        """ Internal method that (re-)constructs the polynomial function.
        """
        coeffs = self._coeffs
        reference, reference2 = self._references
        varNames = self._variables
        
        # Construct the polynomial expression string
        strF = "(%s*x*y)+(%s*x)+(%s*y)+%s"
        src = strF % tuple(map(self._floatOrInt, coeffs))

        # Build the display (i.e. unoptimized) version of the polynomial         
        self._str = strF % self._coeffs
        varNames = list(varNames)
        if reference != 0:
            varNames[0] = "(x-%s)" % reference
        if reference2 != 0:
            varNames[1] = "(y-%s)" % reference2
        self._str = self._fixSums(self._str)
        
        # Replace standard variable names with custom ones. Change one to 
        # a stand-in first, to prevent problems if one name contains the other.
        self._str = self._streplace(self._str, ("x", "\x00"), 
                                    ("y", varNames[1]), ("\x00", varNames[0]))

        # Optimizations: Build a simplified expression for function.
        # 1. Remove multiples of 0 and 1, addition of 0 constants.
        src = self._reduce(src)    

        # 2. If there's a reference value, replace the variable with (v-ref)
        references = [self._floatOrInt(x) for x in (reference, reference2)]
        for i, v in enumerate("xy"):
            if references[i] != 0:
                src = src.replace(v, "(%s-%s)" % (v, references[i]))  
        
        # 3. Do the reduction again, now that the offsets are in.
        src = self._reduce(src)    
        if not src:
            logger.warning("Bad polynomial coefficients {}".format(self._coeffs))
            src = "x"
        
        self._source = 'lambda x,y: %s' % self._fixSums(src)
        self._function = eval(self._source, {'math': math})
        
        # Optimization: it is possible that the polynomial could exclude Y
        # completely. If that's the case, use a dummy value to speed things up.
        self._noY = (0, 1) if 'y' not in src else False 


    def __call__(self, timestamp, value, session=None, noBivariates=False):
        """ Apply the polynomial to an event. 
        
            :param timestamp: The time of the event to process.
            :param value: The value of the event to process.
            :keyword session: The session containing the event.
            :keyword noBivariates: If `True`, the reference channel will not
                be used.
        """
        session = self.dataset.lastSession if session is None else session
        sessionId = None if session is None else session.sessionId
        
        try:
            if self._eventlist is None or self._sessionId != sessionId:
                channel = self.dataset.channels[self.channelId][self.subchannelId]
                self._eventlist = channel.getSession(session.sessionId)
                self._sessionId = session.sessionId
            if len(self._eventlist) == 0:
                return timestamp, value
            
            # Optimization: don't check the other channel if Y is unused
            if noBivariates:
                y = (0, 1)
            else:
                y = self._noY or self._eventlist.getMeanNear(timestamp)
            return timestamp, self._function(value, y)
        
        except (IndexError, ZeroDivisionError) as err:
            # In multithreaded environments, there's a rare race condition
            # in which the main channel can be accessed before the calibration
            # channel has loaded. This should fix it.
            logger.warning("%s occurred in Bivariate polynomial %r" %
                           err.__class__.__name__, self.id)
            return None


    def asDict(self):
        """ Dump the polynomial as a dictionary. Intended for use when
            generating EBML.
        """
        cal = super(Bivariate, self).asDict()
        cal['BivariateCalReferenceValue'] = self._references[1]
        cal['BivariateChannelIDRef'] = self.channelId
        cal['BivariateSubChannelIDRef'] = self.subchannelId
        return cal


    def isValid(self, session=None, noBivariates=False, _retries=3):
        """ Check the validity of the Transform.
        
            :keyword session: The session to check (could be valid in one and
                invalid in another, e.g. one session has no temperature data).
            :keyword noBivariates: If `True`, the reference channel will not
                be used.
        """
        valid = super(Bivariate, self).isValid(session, noBivariates)
        if noBivariates or not valid:
            return valid
        
        session = self.dataset.lastSession if session is None else session
        sessionId = None if session is None else session.sessionId
        
        try:
            if self._eventlist is None or self._sessionId != sessionId:
                channel = self.dataset.channels[self.channelId][self.subchannelId]
                self._eventlist = channel.getSession(session.sessionId)
                self._sessionId = session.sessionId
            if len(self._eventlist) == 0:
                return False
            
        except:
            # HACK: In multithreaded environments, there's a rare race 
            # condition in which the main channel can be accessed before the
            # calibration channel has loaded. Retry isValid() a few times.
            if _retries == 0:
                return False
            
            return self.isValid(session, noBivariates, _retries-1)


#===============================================================================
# 
#===============================================================================

class CombinedPoly(Bivariate):
    """ Calibration transform that combines multiple polynomials into a single
        function. Used for combining Channel and Subchannel transforms to make
        them more efficient.
    """
    
    def copy(self):
        """ Create a duplicate of this Transform.
        """
        return self.__class__(self.poly, subchannel=self._subchannel, 
                              calId=self.id, dataset=self.dataset, 
                              **self.kwargs)
    
    def __init__(self, poly, subchannel=None, calId=None, dataset=None, 
                 **kwargs):
        self.id = calId
        self.poly = poly
        self.subpolys = kwargs
        
        if subchannel is not None:
            self.poly = self.poly[subchannel]

        self.kwargs = kwargs
        p = list(kwargs.values())[0]
        for attr in ('dataset', '_eventlist', '_sessionId', 'channelId',
                     'subchannelId', '_references', '_coeffs', '_variables'):
            setattr(self, attr, getattr(poly, attr, getattr(p, attr, None)))
                    
        self.dataset = self.dataset or dataset
        self._subchannel = subchannel
        self._build()
    
    
    def _build(self):
        if self.poly is None:
            if len(self.subpolys) == 1:
                p = list(self.subpolys.values())[0]
                if p is not None:
                    for attr in ('_str', '_source', '_function', '_noY'):
                        setattr(self, attr, getattr(p, attr, None))
                    return
            phead, src = "lambda x", "x"
        else:
            phead, src = self.poly.source.split(": ")
            
        for k, v in self.subpolys.items():
            if v is None:
                ssrc = "lambda x: x"
            else:
                ssrc = v.source 
            s = "(%s)" % ssrc.split(": ")[-1]
            src = self._reduce(src.replace(k, s))

        if self._subchannel is not None:
            src = src.replace('x', 'x[%d]' % self._subchannel)
        
        # Merge in all function globals, in case components use additional
        # libraries (e.g. math). 
        evalGlobals = {'math': math}
        if self.poly is not None: 
            evalGlobals.update(self.poly._function.__globals__)
        for p in self.subpolys.values():
            if p is not None:
                evalGlobals.update(p._function.__globals__)
        
        self._str = src
        self._source = "%s: %s" % (phead, src)
        self._function = eval(self._source, evalGlobals)
        self._noY = (0, 1) if 'y' not in src else False
        
        
    def asDict(self):
        """ Dump the polynomial as a dictionary. Intended for use when
            generating EBML.
        """
        raise TypeError("Can't generate dictionary for %s" %
                        self.__class__.__name__)

    
    def isValid(self, session=None, noBivariates=False):
        """ Check the validity of the Transform.
        """
        if not Transform.isValid(self, session, noBivariates):
            return False
        return all(p.isValid(session, noBivariates) for p in self.subpolys.values())
        

#------------------------------------------------------------------------------ 

class PolyPoly(CombinedPoly):
    """ Calibration transform that combines multiple subchannel polynomials 
        into one single function.
    """
    
    def copy(self):
        """ Create a duplicate of this Transform.
        """
        return self.__class__(self.polys, calId=self.id, dataset=self.dataset)
    
    
    def __init__(self, polys, calId=None, dataset=None):
        self.id = calId
        self.polys = polys
        poly = polys[0]
        for attr in ('dataset', '_eventlist', '_sessionId', 'channelId', 'subchannelId'):
            setattr(self, attr, getattr(poly, attr, None))
        
        self.dataset = self.dataset or dataset
        self._eventlist = None
        self._build()


    def _build(self):
        params = []
        body = []
        for n, p in enumerate(self.polys):
            params.append('x%d' % n)
            if p is None:
                body.append('x%d' % n)
            else:
                body.append(p.source.split(':')[-1].replace('x', 'x%d' % n))
        
        # ends with a comma to ensure a tuple is returned
        src = "(%s,)" % (', '.join(body))
        self._str = src

        if 'y' in src:
            params.insert(0, 'y')
            self._noY = False
        else:
            self._noY = (0, 1)
        
        # Merge in all function globals, in case components use additional
        # libraries (e.g. math). 
        evalGlobals = {'math': math}
        for p in self.polys:
            if p is not None:
                evalGlobals.update(p._function.__globals__)
        
        self._source = "lambda %s: %s" % (','.join(params), src)
        self._function = eval(self._source, evalGlobals)
        self._variables = params


    def __call__(self, timestamp, values, session=None, noBivariates=False):
        """ Apply the polynomial to an event. 
        
            :param timestamp: The time of the event to process.
            :param values: The values of the event to process.
            :keyword session: The session containing the event.
        """
        try:
            # Optimization: don't check the other channel if Y is unused
            if self._noY is False:
                if noBivariates:
                    return timestamp, self._function(0, *values)
                    
                session = self.dataset.lastSession if session is None else session
                sessionId = None if session is None else session.sessionId
                
                if self._eventlist is None or self._sessionId != sessionId:
                    channel = self.dataset.channels[self.channelId][self.subchannelId]
                    self._eventlist = channel.getSession(session.sessionId)
                    self._sessionId = session.sessionId
                
                # XXX: Hack! EventList length can be 0 if a thread is running.
                # This almost immediately gets fixed. Find real cause.
                try:    
                    y = self._eventlist.getMeanNear(timestamp, outOfRange=True)
                except IndexError:
                    sleep(0.001)
                    if len(self._eventlist) == 0:
                        return None
                    y = self._eventlist.getMeanNear(timestamp, outOfRange=True)
                    
                return timestamp, self._function(y, *values)
            
            else:
                return timestamp, self._function(*values)
            
        except (TypeError, IndexError, ZeroDivisionError) as err:
            # In multithreaded environments, there's a rare race condition
            # in which the main channel can be accessed before the calibration
            # channel has loaded. This should fix it.
            if getattr(self.dataset, 'loading', False):
                logger.warning("%s occurred in combined polynomial %r" %
                               (err.__class__.__name__, self))
                return None
            raise


    def isValid(self, session=None, noBivariates=False):
        """ Check the validity of the Transform.
        """
        if not Transform.isValid(self, session, noBivariates):
            return False
        return all(p.isValid(session, noBivariates) for p in self.polys)



#!
"""
Package for reading and analyzing Mide Instrumentation Data Exchange (MIDE)
files.
"""

__author__ = "David Randall Stokes"
__copyright__ = "Copyright (c) 2020 Mid\xe9 Technology"

__maintainer__ = "David Randall Stokes"
__email__ = "dstokes@mide.com"

__version__ = (3, 0, 0)
__status__ = "Development"

from .importer import importFile



"""
Data type conversion. Each class contains two special attributes:

* conversion: A tuple containing a pair of tuples, the label and units of the
    'from' and 'to'.
* parameters: If the converter has additional parameters, these are listed
    here.

:author: dstokes
"""
import math

import numpy as np

from .transforms import Univariate
from .dataset import Channel

# ==============================================================================
# 
# ==============================================================================

CONVERTERS = []


def registerConverter(cls):
    """ Decorator. Used to register classes as parsers of data payloads. 
    """
    global CONVERTERS
    CONVERTERS.append(cls)
    return cls


# ==============================================================================
# 
# ==============================================================================

class UnitConverter(Univariate):
    """ Mix-in class for unit conversion transforms.
    """
    modifiesValue = True
    modifiesTime = False
    parameters = None
    
    @classmethod
    def isApplicable(cls, obj):
        """ Is this converter applicable to the given object?
         
            :param obj: A `Channel` or `EventList` (or a subclass of either).
                Can also be a list or tuple of said objects. In the latter case,
                the applicability of each item is tested, and `True` returned
                if they are all applicable.
        """
        if isinstance(obj, (list, tuple)):
            if len(obj) == 0:
                return False
            return all(map(cls.isApplicable, obj))
        if isinstance(obj, Channel):
            sourceUnits = obj.units
        else:
            sourceUnits = obj.parent.units
        fromUnits = cls.convertsFrom
        if sourceUnits is None:
            return False
        if sourceUnits == fromUnits:
            return True
        if None not in cls.convertsFrom:
            return False
        units = (fromUnits[0] or sourceUnits[0], fromUnits[1] or sourceUnits[1])
        return units == sourceUnits
    
    def convert(self, v):
        """ Convert a value to new units.  Aliases self.function() """
        return self.function(v)
    
    def revert(self, v):
        """ Convert a value back to the original units. Primarily for display
            purposes.
        """
        # TODO: Make this work on Univariates with more than 2 coefficients.
        # May never be needed.
        a, b = self.coefficients
        ref = self.references[0]
        return ((v-b)/a)+ref

# ==============================================================================
# Simple converters
# ==============================================================================


@registerConverter
class Celsius2Fahrenheit(UnitConverter):
    """ Convert degrees Celsius to Fahrenheit. 
    """
    convertsFrom = ('Temperature', '\xb0C')
    units = ('Temperature', '\xb0F')
    
    def __init__(self, calId=None, dataset=None, varName="x"):
        """
        """
        super(Celsius2Fahrenheit, self).__init__((1.8, 32), calId=calId,
                                                 dataset=dataset, varName=varName)
    

@registerConverter
class Celsius2Kelvin(UnitConverter):
    """ Convert degrees Celsius to Kelvin. 
    """
    convertsFrom = ('Temperature', '\xb0C')
    units = ('Temperature', '\xb0K')

    def __init__(self, calId=None, dataset=None, varName="x"):
        """
        """
        super(Celsius2Kelvin, self).__init__((1, 273.15), calId=calId,
                                             dataset=dataset, varName=varName)
        

@registerConverter
class Gravity2MPerSec2(UnitConverter):
    """ Convert acceleration from g to m/s^2.
    """
    convertsFrom = ('Acceleration', 'g')
    units = ('Acceleration', 'm/s\u00b2')
    
    def __init__(self, calId=None, dataset=None, varName='x'):
        super(Gravity2MPerSec2, self).__init__((9.806649999788, 0), calId=calId,
                                               dataset=dataset, varName=varName)


@registerConverter
class Meters2Feet(UnitConverter):
    """ Convert meters to feet.
    """
    convertsFrom = (None, 'm')
    units = (None, 'ft')
    
    def __init__(self, calId=None, dataset=None, varName='x'):
        super(Meters2Feet, self).__init__((3.2808399, 0), calId=calId, 
                                          dataset=dataset, varName=varName)


@registerConverter
class Pa2PSI(UnitConverter):
    """ Convert air pressure from Pascals to pounds per square inch.
    """
    convertsFrom = ('Pressure', 'Pa')
    units = ('Pressure', 'psi')
    
    def __init__(self, calId=None, dataset=None, varName='x'):
        super(Pa2PSI, self).__init__((0.000145037738, 0), calId=calId,
                                     dataset=dataset, varName=varName)


@registerConverter
class Pa2atm(UnitConverter):
    """ Convert air pressure from Pascals to atmospheres.
    """
    convertsFrom = ('Pressure', 'Pa')
    units = ('Pressure', 'atm')
    # 9.86923266716e-06
    
    def __init__(self, calId=None, dataset=None, varName='x'):
        super(Pa2atm, self).__init__((9.86923266716e-06, 0), calId=calId,
                                     dataset=dataset, varName=varName)

# ==============================================================================
# More complex conversion
# ==============================================================================


@registerConverter
class Pressure2Meters(UnitConverter):
    """ Convert pressure in Pascals to an altitude in meters.
    """
    convertsFrom = ('Pressure', 'Pa')
    units = ('Altitude', 'm')
    
    # Parameters: name, description, type, range, and default value.
    # The names must match keyword arguments in __init__ and object attributes.
    parameters = (('temp', 'Temperature at sea level (\xb0C)', float, (-100, 100), 15.0),
                  ('sealevel', 'Pressure at sea level (Pa)', int, (0, 150000), 101325.0))
    
    def copy(self):
        """ Create a duplicate of this unit converter.
        """
        t = self.__class__(self.id, self.dataset, self._temp, self._sealevel)
        return t
    
    def __init__(self, calId=None, dataset=None, temp=15.0, sealevel=101325.0):
        """ Constructor.
            :keyword dataset: The `Dataset` to which this applies. 
            :keyword temp: Temperature at sea level (degrees C)
            :keyword sealevel: Air pressure at sea level (Pascals)
        """
        self._sealevel = sealevel
        self._temp = temp
        self._tempK = temp + 273.15
        self.id = calId
        self._lastSession = None
        self._timeOffset = 0
        self._build()

    def __hash__(self):
        return hash((self.__class__, self._temp, self._sealevel))

    def _build(self):
        # This is sort of a special case; the function is too complex to
        # nicely represent as a lambda, so the 'source' calls the object itself.
        # This is so the polynomial combination/reduction will work.
        # `_function` is never actually called; `function` is overridden.
        self.T_2 = self._tempK - 71.5
        self.h_1 = ((8.31432*self.T_2*(math.log(101325/22632.1)))/((-9.80665)*0.0289644))
        self._str = "Pressure2Meters.convert(x)"
        self._source = "lambda x: %s" % self._str
        self._function = eval(self._source, 
                              {'Pressure2Meters': self, 'math': math, 'np': np})

    @property
    def sealevel(self):
        return self._sealevel
    
    @sealevel.setter
    def sealevel(self, p):
        self._sealevel = float(p)
        self._build()
        
    @property
    def temp(self):
        return self._temp
    
    @temp.setter
    def temp(self, t):
        self._temp = t
        self._tempK = t + 273.15
        self._build()

    def function(self, p):
        p = np.asarray(p)
        result = np.full_like(p, fill_value=20e3)  # filled with default value

        sp = self._sealevel / p
        mask = (sp < 4.47704808656731)
        if mask.any():
            foo = np.power((p[mask]/self._sealevel), -0.1902632365084836)
            result[mask] = (self._tempK*((1.0/foo)-1.0))/-0.0065

        mask = (sp < 18.507221149648668) & ~mask
        if mask.any():
            T_2 = self._tempK - 71.5
#             h_2 = (8.31432*T_2*(math.log(p/self._sealevel)))/-0.28404373326
#             h_1 = ((T_2*12.462865699354536)/-0.28404373326)+11000
            h_2 = (T_2*np.log(p[mask]/self._sealevel))/-0.03416319473631036
            h_1 = (T_2/-0.02279120549896569)+11000.0
            result[mask] = h_1+h_2

        return result

    def revert(self, h):
        h = np.asarray(h)
        result = np.full_like(h, fill_value=5474.89)  # filled with default value

        M = 0.0289644  # [kg/mol] molar mass of Earth's air
        g = 9.80665  # [m/s^2] gravitational acceleration constant
        R = 8.31432  # [(N*m)/(mol*k)] universal gas constant

        t = self._tempK
        p_a = self._sealevel

        mask = (h < 11000)
        if mask.any():
            # [K/m] temperature lapse rate
            L_a = -0.0065
            # [m] height above sea level (differing altitudes have differing time lapse rates)
            h_a = 0.0
            result[mask] = self._sealevel*np.power(
                self._tempK / (self._tempK + L_a*(h[mask]-h_a)),
                (g*M)/(R*L_a)
                )

        mask = (h <= 20000) & ~mask
        if mask.any():
            L_a = -0.0065
            h_a = 0.0
            h_b = 11000
            p_b = p_a*np.power(t/(t+(L_a*(h_b-h_a))), (g*M)/(R*L_a))
            T_1 = t+(11000*(-0.0065))
            result[mask] = p_b*np.exp((-g*M*(h[mask]-h_b))/(R*T_1))

        return result


@registerConverter
class Pressure2Feet(Pressure2Meters):
    """ Convert pressure in Pascals to an altitude in feet.
    """
    convertsFrom = ('Pressure', 'Pa')
    units = ('Altitude', 'ft')
    
    def _build(self):
        super(Pressure2Feet, self)._build()
        self._str = "Pressure2Feet.convert(x)"
        self._source = "lambda x: %s" % self._str
        self._function = eval(self._source, 
                              {'Pressure2Feet': self, 'math': math, 'np': np})

    def function(self, press):
        return 3.2808399*Pressure2Meters.function(self, press)

    def revert(self, v):
        return Pressure2Meters.revert(self, v/3.2808399)

# ==============================================================================
# 
# ==============================================================================


def getApplicableConverters(doc):
    """ Get all unit converters applicable to any part of the given Dataset.
    
        :param doc: a `Dataset` instance.
        :return: A list of `UnitConverter` classes.
    """
    global CONVERTERS
    results = {}
    for p in doc.getPlots():
        for c in CONVERTERS:
            if c.isApplicable(p):
                results[c] = True
    return list(results.keys())



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

INSTALL_REQUIRES = [
    'numpy',
    'ebmlite>=3.0.0'
    ]

TEST_REQUIRES = [
    'pytest>=4.6',
    'mock',
    'pytest-cov',
    ]

EXAMPLE_REQUIRES = [
    'matplotlib'
    ]

setuptools.setup(
        name='idelib',
        version='3.1.0',
        author='Mide Technology',
        author_email='help@mide.com',
        description='Python API for accessing IDE data recordings',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/MideTechnology/idelib',
        license='MIT',
        classifiers=['Development Status :: 5 - Production/Stable',
                     'License :: OSI Approved :: MIT License',
                     'Natural Language :: English',
                     'Programming Language :: Python :: 3.5',
                     'Programming Language :: Python :: 3.6',
                     'Programming Language :: Python :: 3.7',
                     'Programming Language :: Python :: 3.8',
                     'Programming Language :: Python :: 3.9',
                     'Topic :: Scientific/Engineering',
                     ],
        keywords='ebml binary ide mide',
        packages=setuptools.find_packages(exclude=('testing',)),
        package_dir={'idelib': './idelib'},
        package_data={
            'idelib': ['schemata/*'],
        },
        test_suite='./testing',
        install_requires=INSTALL_REQUIRES,
        extras_require={
            'test': INSTALL_REQUIRES + TEST_REQUIRES,
            'example': INSTALL_REQUIRES + EXAMPLE_REQUIRES,
            },
)
